from __future__ import absolute_import
from builtins import str
import os, sys, shutil
from .__init__ import ForceBalanceTestCase
from forcebalance.parser import parse_inputs
from forcebalance.forcefield import FF
from forcebalance.objective import Objective
from forcebalance.optimizer import Optimizer, Counter
import pytest

class TestWaterTutorial(ForceBalanceTestCase):
    def setup_method(self, method):
        super(TestWaterTutorial, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        # copy folder 'files/test_liquid' into a new folder 'files/test_liquid.run'
        files_folder = os.path.join(self.cwd, 'files')
        tmpfolder = os.path.join(self.cwd, 'files', 'test_continue.run')
        source_folder = os.path.join(self.cwd, '../../studies/001_water_tutorial')

        if os.path.isdir(tmpfolder):
            shutil.rmtree(tmpfolder)
        os.mkdir(tmpfolder)
        
        shutil.copytree(os.path.join(source_folder, 'forcefield'), os.path.join(tmpfolder, 'forcefield'))
        shutil.copytree(os.path.join(source_folder, 'targets'), os.path.join(tmpfolder, 'targets'))
        shutil.copy(os.path.join(files_folder, "test_continue.sav"), tmpfolder)
        shutil.copy(os.path.join(files_folder, "test_continue.in"), tmpfolder)
        shutil.copytree(os.path.join(files_folder, "test_continue.tmp"), os.path.join(tmpfolder,"test_continue.tmp"))
        os.chdir(tmpfolder)

    def teardown_method(self):
        tmpfolder = os.path.join(self.cwd, 'files', 'test_continue.run')
        if os.path.isdir(tmpfolder):
            shutil.rmtree(tmpfolder)
        super(TestWaterTutorial, self).teardown_method()

    def test_continue(self):
        """Check continuation from a previous run"""
        if sys.version_info < (3,0):
            pytest.skip("Existing pickle file only works with Python 3")
        self.logger.debug("\nSetting input file to 'test_continue.in'\n")
        input_file='test_continue.in'

        ## The general options and target options that come from parsing the input file
        self.logger.debug("Parsing inputs...\n")
        options, tgt_opts = parse_inputs(input_file)
        options['continue'] = True
        self.logger.debug("options:\n%s\n\ntgt_opts:\n%s\n\n" % (str(options), str(tgt_opts)))

        assert isinstance(options, dict), "Parser gave incorrect type for options"
        assert isinstance(tgt_opts, list), "Parser gave incorrect type for tgt_opts"
        for target in tgt_opts:
            assert isinstance(target, dict), "Parser gave incorrect type for target dict"

        ## The force field component of the project
        forcefield  = FF(options)
        assert isinstance(forcefield, FF), "Expected forcebalance forcefield object"

        ## The objective function
        objective   = Objective(options, tgt_opts, forcefield)
        assert isinstance(objective, Objective), "Expected forcebalance objective object"

        ## The optimizer component of the project
        self.logger.debug("Creating optimizer: ")
        optimizer   = Optimizer(options, objective, forcefield)
        assert isinstance(optimizer, Optimizer), "Expected forcebalance optimizer object"
        self.logger.debug(str(optimizer)+'\n')

        ## Actually run the optimizer.
        self.logger.debug("Done setting up! Running optimizer...\n")
        result = optimizer.Run()
        self.logger.debug("\nOptimizer finished. Final results:\n")
        self.logger.debug(str(result)+'\n')

        assert optimizer.iterinit == 2, "Initial iteration counter is incorrect"
        assert optimizer.iteration == 2, "Final iteration counter is incorrect"

        # self.assertNdArrayEqual(EXPECTED_WATER_RESULTS,result,delta=0.001,
        #                         msg="\nCalculation results have changed from previously calculated values.\n"
        #                         "If this seems reasonable, update EXPECTED_WATER_RESULTS in test_system.py with these values")
        # # Fail if calculation takes longer than previously to converge
        # self.assertGreaterEqual(ITERATIONS_TO_CONVERGE, Counter(), msg="\nCalculation took longer than expected to converge (%d iterations vs previous of %d)" %\
        # (ITERATIONS_TO_CONVERGE, Counter()))




from __future__ import absolute_import
import forcebalance
import os, sys, shutil
import tarfile
import logging
import pytest
from .__init__ import ForceBalanceTestCase

logger = logging.getLogger("test")

class TestOptimizer(ForceBalanceTestCase):
    def setup_method(self, method):
        super(TestOptimizer, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(self.cwd, '..', '..', 'studies', '001_water_tutorial'))
        self.input_file = 'very_simple.in'
        targets = tarfile.open('targets.tar.bz2','r')
        targets.extractall()
        targets.close()

        self.options, self.tgt_opts = forcebalance.parser.parse_inputs(self.input_file)

        self.options.update({'writechk':'checkfile.tmp'})

        self.forcefield = forcebalance.forcefield.FF(self.options)
        self.objective = forcebalance.objective.Objective(self.options, self.tgt_opts, self.forcefield)
        try: self.optimizer = forcebalance.optimizer.Optimizer(self.options, self.objective, self.forcefield)
        except: pytest.fail("Couldn't create optimizer")

    def teardown_method(self):
        shutil.rmtree("results", ignore_errors=True)
        shutil.rmtree("*.bak", ignore_errors=True)
        shutil.rmtree("*.tmp", ignore_errors=True)
        super(TestOptimizer, self).teardown_method()

    def test_optimizer(self):
        self.optimizer.writechk()
        assert os.path.isfile(self.options['writechk']), \
            "Optimizer.writechk() didn't create expected file at %s " % self.options['writechk']
        read = self.optimizer.readchk()
        assert isinstance(read, dict)



""" @package forcebalance.opt_geo_target Optimized Geometry fitting module.

@author Yudong Qiu, Lee-Ping Wang
@date 03/2019
"""
from __future__ import division
import os
import shutil
import numpy as np
import re
import subprocess
from copy import deepcopy
from collections import OrderedDict, defaultdict
from forcebalance.nifty import col, eqcgmx, flat, floatornan, fqcgmx, invert_svd, kb, printcool, printcool_dictionary, bohr2ang, warn_press_key
from forcebalance.target import Target
from forcebalance.molecule import Molecule
from forcebalance.finite_difference import fdwrap, f1d2p, f12d3p, in_fd
from forcebalance.output import getLogger
logger = getLogger(__name__)

RADIAN_2_DEGREE = 180 / np.pi

def periodic_diff(a, b, v_periodic):
    ''' convenient function for computing the minimum difference in periodic coordinates
    Parameters
    ----------
    a: np.ndarray or float
        Reference values in a numpy array
    b: np.ndarray or float
        Target values in a numpy arrary
    v_periodic: float > 0
        Value of the periodic boundary

    Returns
    -------
    diff: np.ndarray
        The array of same shape containing the difference between a and b
        All return values are in range [-v_periodic/2, v_periodic/2),
        "( )" means exclusive, "[ ]" means inclusive

    Examples
    -------
    periodic_diff(0.0, 2.1, 2.0) => -0.1
    periodic_diff(0.0, 1.9, 2.0) => 0.1
    periodic_diff(0.0, 1.0, 2.0) => -1.0
    periodic_diff(1.0, 0.0, 2.0) => -1.0
    periodic_diff(1.0, 0.1, 2.0) => -0.9
    periodic_diff(1.0, 10.1, 2.0) => 0.9
    periodic_diff(1.0, 9.9, 2.0) => -0.9
    '''
    assert v_periodic > 0
    h = 0.5 * v_periodic
    return (a - b + h) % v_periodic - h

def compute_rmsd(ref, tar, v_periodic=None):
    """
    Compute the RMSD between two arrays, supporting periodic difference
    """
    assert len(ref) == len(tar), 'array length must match'
    n = len(ref)
    if n == 0: return 0.0
    if v_periodic is not None:
        diff = periodic_diff(ref, tar, v_periodic)
    else:
        diff = ref - tar
    rmsd = np.sqrt(np.sum(diff**2) / n)
    return rmsd

class OptGeoTarget(Target):
    """ Subclass of Target for fitting MM optimized geometries to QM optimized geometries. """
    def __init__(self,options,tgt_opts,forcefield):
        super(OptGeoTarget,self).__init__(options,tgt_opts,forcefield)
        self.set_option(None, None, 'optgeo_options', os.path.join(self.tgtdir,tgt_opts['optgeo_options_txt']))
        self.sys_opts = self.parse_optgeo_options(self.optgeo_options)
        ## Build keyword dictionaries to pass to engine.
        engine_args = OrderedDict(list(self.OptionDict.items()) + list(options.items()))
        engine_args.pop('name', None)
        ## Create engine objects
        self.create_engines(engine_args)
        ## Create internal coordinates
        self._build_internal_coordinates()
        ## Option for how much data to write to disk.
        self.set_option(tgt_opts,'writelevel','writelevel')

    def create_engines(self, engine_args):
        raise NotImplementedError("create_engines() should be implemented in subclass")

    @staticmethod
    def parse_optgeo_options(filename):
        """ Parse an optgeo_options.txt file into specific OptGeoTarget Target Options"""
        logger.info("Reading optgeo options from file: %s\n" % filename)
        global_opts = OrderedDict()
        sys_opts = OrderedDict()
        section = None
        section_opts = OrderedDict()
        with open(filename) as f:
            for ln, line in enumerate(f, 1):
                # Anything after "#" is a comment
                line = line.split("#", maxsplit=1)[0].strip()
                if not line: continue
                ls = line.split()
                key = ls[0].lower()
                if key[0] == "$":
                    # section sign $
                    if key == '$end':
                        if section is None:
                            warn_press_key("Line %i: Encountered $end before any section." % ln)
                        elif section == 'global':
                            # global options read finish
                            global_opts = section_opts
                        elif section == 'system':
                            # check if system section contains name
                            if 'name' not in section_opts:
                                warn_press_key("Line %i: You need to specify a name for the system section ending." % ln)
                            elif section_opts['name'] in sys_opts:
                                warn_press_key("Line %i: A system named %s already exists in Systems" % (ln, section_opts['name']))
                            else:
                                sys_opts[section_opts['name']] = section_opts
                        section = None
                        section_opts = OrderedDict()
                    else:
                        if section is not None:
                            warn_press_key("Line %i: Encountered section start %s before previous section $end." % (ln, key))
                        if key == '$global':
                            section = 'global'
                        elif key == '$system':
                            section = 'system'
                        else:
                            warn_press_key("Line %i: Encountered unsupported section name %s " % (ln, key))
                else:
                    # put normal key-value options into section_opts
                    if key in ['name', 'geometry', 'topology']:
                        if len(ls) != 2:
                            warn_press_key("Line %i: one value expected for key %s" % (ln, key))
                        if section == 'global':
                            warn_press_key("Line %i: key %s should not appear in $global section" % (ln, key))
                        section_opts[key] = ls[1]
                    elif key in ['bond_denom', 'angle_denom', 'dihedral_denom', 'improper_denom']:
                        if len(ls) != 2:
                            warn_press_key("Line %i: one value expected for key %s" % (ln, key))
                        section_opts[key] = float(ls[1])
                    elif key == 'mol2':
                        # special parsing for mol2 option for SMIRNOFF engine
                        # the value is a list of filenames
                        section_opts[key] = ls[1:]
        # apply a few default global options
        global_opts.setdefault('bond_denom', 0.02)
        global_opts.setdefault('angle_denom', 3)
        global_opts.setdefault('dihedral_denom', 10.0)
        global_opts.setdefault('improper_denom', 10.0)
        # copy global options into each system
        for sys_name, sys_opt_dict in sys_opts.items():
            for k,v in global_opts.items():
                # do not overwrite system options
                sys_opt_dict.setdefault(k, v)
            for k in ['name', 'geometry', 'topology']:
                if k not in sys_opt_dict:
                    warn_press_key("key %s missing in system section named %s" %(k, sys_name))
        return sys_opts

    def _build_internal_coordinates(self):
        "Build internal coordinates system with geometric.internal.PrimitiveInternalCoordinates"
        # geometric module is imported to build internal coordinates
        # importing here will avoid import error for calculations not using this target
        from geometric.internal import PrimitiveInternalCoordinates, Distance, Angle, Dihedral, OutOfPlane
        self.internal_coordinates = OrderedDict()
        for sysname, sysopt in self.sys_opts.items():
            geofile = os.path.join(self.root, self.tgtdir, sysopt['geometry'])
            topfile = os.path.join(self.root, self.tgtdir, sysopt['topology'])
            # logger.info("Building internal coordinates from file: %s\n" % topfile)
            m0 = Molecule(geofile)
            m = Molecule(topfile)
            p_IC = PrimitiveInternalCoordinates(m)
            # here we explicitly pick the bonds, angles and dihedrals to evaluate
            ic_bonds, ic_angles, ic_dihedrals, ic_impropers = [], [], [], []
            for ic in p_IC.Internals:
                if isinstance(ic, Distance):
                    ic_bonds.append(ic)
                elif isinstance(ic, Angle):
                    ic_angles.append(ic)
                elif isinstance(ic, Dihedral):
                    ic_dihedrals.append(ic)
                elif isinstance(ic, OutOfPlane):
                    ic_impropers.append(ic)
            # compute and store reference values
            pos_ref = m0.xyzs[0]
            vref_bonds = np.array([ic.value(pos_ref) for ic in ic_bonds])
            vref_angles = np.array([ic.value(pos_ref)*RADIAN_2_DEGREE for ic in ic_angles])
            vref_dihedrals = np.array([ic.value(pos_ref)*RADIAN_2_DEGREE for ic in ic_dihedrals])
            vref_impropers = np.array([ic.value(pos_ref)*RADIAN_2_DEGREE for ic in ic_impropers])
            self.internal_coordinates[sysname] = {
                'ic_bonds': ic_bonds,
                'ic_angles': ic_angles,
                'ic_dihedrals': ic_dihedrals,
                'ic_impropers': ic_impropers,
                'vref_bonds': vref_bonds,
                'vref_angles': vref_angles,
                'vref_dihedrals': vref_dihedrals,
                'vref_impropers': vref_impropers,
            }

    def system_driver(self, sysname, save_mol=None):
        """ Run calculation for one system, return internal coordinate values after optimization """
        engine = self.engines[sysname]
        ic_dict = self.internal_coordinates[sysname]
        if engine.__class__.__name__ in ('OpenMM', 'SMIRNOFF'):
            # OpenMM.optimize() by default resets geometry to initial geometry before optimization
            engine.optimize(0)
            pos = engine.getContextPosition()
            if save_mol is not None:
                MM_minimized_mol = deepcopy(engine.mol[0])
                MM_minimized_mol.xyzs[0] = pos
                MM_minimized_mol.write(save_mol)
        else:
            raise NotImplementedError("system_driver() not implemented for %s" % engine.__name__)
        v_ic = {
            'bonds': np.array([ic.value(pos) for ic in ic_dict['ic_bonds']]),
            'angles': np.array([ic.value(pos)*RADIAN_2_DEGREE for ic in ic_dict['ic_angles']]),
            'dihedrals': np.array([ic.value(pos)*RADIAN_2_DEGREE for ic in ic_dict['ic_dihedrals']]),
            'impropers': np.array([ic.value(pos)*RADIAN_2_DEGREE for ic in ic_dict['ic_impropers']]),
        }
        return v_ic

    def indicate(self):
        title_str = "%s, Objective = % .5e" % (self.name, self.objective)
        #QYD: This title is carefully placed to align correctly
        column_head_str1 =  " %-20s %13s     %13s     %15s   %15s   %17s " % ("System", "Bonds", "Angles", "Dihedrals", "Impropers", "Term.")
        column_head_str2 =  " %-20s %9s %7s %9s %7s %9s %7s %9s %7s %17s " % ('', 'RMSD', 'denom', 'RMSD', 'denom', 'RMSD', 'denom', 'RMSD', 'denom', '')
        printcool_dictionary(self.PrintDict,title=title_str + '\n' + column_head_str1 + '\n' + column_head_str2, center=[True,False,False])

    def get(self, mvals, AGrad=False, AHess=False):
        Answer = {'X':0.0, 'G':np.zeros(self.FF.np), 'H':np.zeros((self.FF.np, self.FF.np))}
        self.PrintDict = OrderedDict()
        # enable self.system_mval_masks (supported by OptGeoTarget_SMIRNOFF)
        enable_system_mval_mask = hasattr(self, 'system_mval_masks')
        def compute(mvals, p_idx=None):
            ''' Compute total objective value for each system '''
            self.FF.make(mvals)
            v_obj_list = []
            for sysname, sysopt in self.sys_opts.items():
                # ref values of each type
                vref_bonds = self.internal_coordinates[sysname]['vref_bonds']
                vref_angles = self.internal_coordinates[sysname]['vref_angles']
                vref_dihedrals = self.internal_coordinates[sysname]['vref_dihedrals']
                vref_impropers = self.internal_coordinates[sysname]['vref_impropers']
                # counts of each type
                n_bonds = len(vref_bonds)
                n_angles = len(vref_angles)
                n_dihedrals = len(vref_dihedrals)
                n_impropers = len(vref_impropers)
                # use self.system_mval_masks to skip evaluations when computing gradients
                if enable_system_mval_mask and in_fd() and (p_idx is not None) and (self.system_mval_masks[sysname][p_idx] == False):
                    v_obj_list += [0] * (n_bonds + n_angles + n_dihedrals + n_impropers)
                    continue
                # read denominators from system options
                bond_denom = sysopt['bond_denom']
                angle_denom = sysopt['angle_denom']
                dihedral_denom = sysopt['dihedral_denom']
                improper_denom = sysopt['improper_denom']
                # inverse demon to be scaling factors, 0 for denom 0
                scale_bond = 1.0 / bond_denom if bond_denom != 0 else 0.0
                scale_angle = 1.0 / angle_denom if angle_denom != 0 else 0.0
                scale_dihedral = 1.0 / dihedral_denom if dihedral_denom != 0 else 0.0
                scale_improper = 1.0 / improper_denom if improper_denom != 0 else 0.0
                # calculate new internal coordinates
                v_ic = self.system_driver(sysname)
                # objective contribution from bonds
                vtar_bonds = v_ic['bonds']
                diff_bond = ((vref_bonds - vtar_bonds) * scale_bond).tolist() if n_bonds > 0 else []
                # objective contribution from angles
                vtar_angles = v_ic['angles']
                diff_angle = (periodic_diff(vref_angles, vtar_angles, 360) * scale_angle).tolist() if n_angles > 0 else []
                # objective contribution from dihedrals
                vtar_dihedrals = v_ic['dihedrals']
                diff_dihedral = (periodic_diff(vref_dihedrals, vtar_dihedrals, 360) * scale_dihedral).tolist() if n_dihedrals > 0 else []
                # objective contribution from improper dihedrals
                vtar_impropers = v_ic['impropers']
                diff_improper = (periodic_diff(vref_impropers, vtar_impropers, 360) * scale_improper).tolist() if n_impropers > 0 else []
                # combine objective values into a big result list
                sys_obj_list = diff_bond + diff_angle + diff_dihedral + diff_improper
                # extend the result v_obj_list by individual terms in this system
                v_obj_list += sys_obj_list
                # save print string
                if not in_fd():
                    # For printing, we group the RMSD by type
                    rmsd_bond = compute_rmsd(vref_bonds, vtar_bonds)
                    rmsd_angle = compute_rmsd(vref_angles, vtar_angles, v_periodic=360)
                    rmsd_dihedral = compute_rmsd(vref_dihedrals, vtar_dihedrals, v_periodic=360)
                    rmsd_improper = compute_rmsd(vref_impropers, vtar_impropers, v_periodic=360)
                    obj_total = sum(v**2 for v in sys_obj_list)
                    self.PrintDict[sysname] = "% 9.3f % 7.2f % 9.3f % 7.2f % 9.3f % 7.2f % 9.3f % 7.2f %17.3f" % (rmsd_bond, \
                        bond_denom, rmsd_angle, angle_denom, rmsd_dihedral, dihedral_denom, rmsd_improper, improper_denom, obj_total)
            return np.array(v_obj_list, dtype=float)

        V = compute(mvals)
        Answer['X'] = np.dot(V,V)
        # write objective decomposition if wanted
        if self.writelevel > 0:
            # recover mvals
            self.FF.make(mvals)
            with open('rmsd_decomposition.txt', 'w') as fout:
                for sysname in self.internal_coordinates:
                    fout.write("\n[ %s ]\n" % sysname)
                    fout.write('%-25s %15s %15s %15s\n' % ("Internal Coordinate", "Ref QM Value", "Cur MM Value", "Difference"))
                    # reference data
                    sys_data = self.internal_coordinates[sysname]
                    sys_data['ic_bonds']
                    # compute all internal coordinate values again and save mm optimized geometry in xyz file
                    v_ic = self.system_driver(sysname, save_mol='%s_mmopt.xyz' % sysname)
                    for p in ['bonds', 'angles', 'dihedrals', 'impropers']:
                        fout.write('--- ' + p + ' ---\n')
                        ic_list = sys_data['ic_' + p]
                        ref_v = sys_data['vref_' + p]
                        tar_v = v_ic[p]
                        # print each value
                        for ic, v1, v2 in zip(ic_list, ref_v, tar_v):
                            diff = periodic_diff(v1, v2, v_periodic=360) if p != 'bonds' else v1-v2
                            fout.write('%-25s %15.5f %15.5f %+15.3e\n' % (ic, v1, v2, diff))
        # compute gradients and hessian
        dV = np.zeros((self.FF.np,len(V)))
        if AGrad or AHess:
            for p in self.pgrad:
                dV[p,:], _ = f12d3p(fdwrap(compute, mvals, p, p_idx = p), h = self.h, f0 = V)

        for p in self.pgrad:
            Answer['G'][p] = 2*np.dot(V, dV[p,:])
            for q in self.pgrad:
                Answer['H'][p,q] = 2*np.dot(dV[p,:], dV[q,:])
        if not in_fd():
            self.objective = Answer['X']
            self.FF.make(mvals)
        return Answer



from __future__ import absolute_import

from builtins import str
import os, shutil
import tarfile
from .__init__ import ForceBalanceTestCase
from forcebalance.parser import parse_inputs
from forcebalance.forcefield import FF
from forcebalance.objective import Objective
from forcebalance.optimizer import Optimizer, Counter
from numpy import array
import numpy as np
import pytest

# expected results (mvals) taken from previous runs. Update this if it changes and seems reasonable (updated 10/24/13)
#EXPECTED_WATER_RESULTS = array([3.3192e-02, 4.3287e-02, 5.5072e-03, -4.5933e-02, 1.5499e-02, -3.7655e-01, 2.4720e-03, 1.1914e-02, 1.5066e-01])
EXPECTED_WATER_RESULTS = array([4.2370e-02, 3.1217e-02, 5.6925e-03, -4.8114e-02, 1.6735e-02, -4.1722e-01, 6.2716e-03, 4.6306e-03, 2.5960e-01])

# expected results (mvals) taken from previous runs. Update this if it changes and seems reasonable (updated 01/24/14)
EXPECTED_BROMINE_RESULTS = array([-0.305718, -0.12497])

# expected objective function from 003d evaluator bromine study. (updated 11/23/19)
EXPECTED_EVALUATOR_BROMINE_OBJECTIVE = array([1000])

# expected gradient elements from 003d evaluator bromine study. Very large uncertainties of +/- 2000 (updated 11/23/19)
EXPECTED_EVALUATOR_BROMINE_GRADIENT = array([4500, 5500])

# expected result (pvals) taken from ethanol GB parameter optimization. Update this if it changes and seems reasonable (updated 09/05/14)
EXPECTED_ETHANOL_RESULTS = array([1.2286e-01, 8.3624e-01, 1.0014e-01, 8.4533e-01, 1.8740e-01, 6.8820e-01, 1.4606e-01, 8.3518e-01])

# fail test if we take more than this many iterations to converge. Update this as necessary
ITERATIONS_TO_CONVERGE = 5

# expected results taken from previous runs. Update this if it changes and seems reasonable (updated 07/23/14)
EXPECTED_LIPID_RESULTS = array([-6.7553e-03, -2.4070e-02])

# expected results taken from OpenFF torsion profile optimization using OpenFF toolkit 0.4.1 and OpenEye toolkit 2019.10.2. (updated 11/02/19)
EXPECTED_OPENFF_TORSIONPROFILE_RESULTS = array([-9.4238e-02, 7.3350e-03, -7.9467e-05, 1.7172e-02, -1.3309e-01, 6.0076e-02, 1.7895e-02, 6.5866e-02, -1.4084e-01, -2.2906e-02])


class ForceBalanceSystemTest(ForceBalanceTestCase):
    def teardown_method(self):
        for fnm in [self.input_file.replace('.in','.sav')]:
            if os.path.exists(fnm):
                os.remove(fnm)
        for dnm in [self.input_file.replace('.in','.bak'), self.input_file.replace('.in','.tmp'), "result"]:
            if os.path.exists(dnm):
                shutil.rmtree(dnm)
        super(ForceBalanceSystemTest, self).teardown_method()
        
    def get_objective(self):
        """ Return the objective function object """
        ## The general options and target options that come from parsing the input file
        self.logger.debug("Parsing inputs...\n")
        options, tgt_opts = parse_inputs(self.input_file)
        self.logger.debug("options:\n%s\n\ntgt_opts:\n%s\n\n" % (str(options), str(tgt_opts)))
        assert isinstance(options, dict), "Parser gave incorrect type for options"
        assert isinstance(tgt_opts, list), "Parser gave incorrect type for tgt_opts"
        for target in tgt_opts:
            assert isinstance(target, dict), "Parser gave incorrect type for target dict"
        ## The force field component of the project
        forcefield  = FF(options)
        assert isinstance(forcefield, FF), "Expected forcebalance forcefield object"
        ## The objective function
        objective   = Objective(options, tgt_opts, forcefield)
        assert isinstance(objective, Objective), "Expected forcebalance objective object"
        return objective

    def get_optimizer(self):
        """ Return the optimizer object """
        ## The general options and target options that come from parsing the input file
        self.logger.debug("Parsing inputs...\n")
        options, tgt_opts = parse_inputs(self.input_file)
        self.logger.debug("options:\n%s\n\ntgt_opts:\n%s\n\n" % (str(options), str(tgt_opts)))
        assert isinstance(options, dict), "Parser gave incorrect type for options"
        assert isinstance(tgt_opts, list), "Parser gave incorrect type for tgt_opts"
        for target in tgt_opts:
            assert isinstance(target, dict), "Parser gave incorrect type for target dict"
        ## The force field component of the project
        forcefield  = FF(options)
        assert isinstance(forcefield, FF), "Expected forcebalance forcefield object"
        ## The objective function
        objective   = Objective(options, tgt_opts, forcefield)
        assert isinstance(objective, Objective), "Expected forcebalance objective object"
        ## The optimizer component of the project
        self.logger.debug("Creating optimizer: ")
        optimizer   = Optimizer(options, objective, forcefield)
        assert isinstance(optimizer, Optimizer), "Expected forcebalance optimizer object"
        self.logger.debug(str(optimizer) + "\n")
        return optimizer

    def run_optimizer(self, check_result=True, check_iter=True, use_pvals=False):
        optimizer = self.get_optimizer()
        ## Actually run the optimizer.
        self.logger.debug("Done setting up! Running optimizer...\n")
        result = optimizer.Run()
        self.logger.debug("\nOptimizer finished. Final results:\n")
        self.logger.debug(str(result) + '\n')
        ## Convert result to physical values if desired.
        if use_pvals:
            result = optimizer.FF.create_pvals(result)
        if check_result:
            msg = "\nCalculation results have changed from previously calculated values.\n " \
                  "If this seems reasonable, update %s in test_system.py with these values" % self.expected_results_name
            np.testing.assert_allclose(self.expected_results, result, atol=self.absolute_tolerance, err_msg=msg)
        if check_iter:
            # Fail if calculation takes longer than previously to converge
            assert ITERATIONS_TO_CONVERGE >= Counter(), "Calculation took longer than expected to converge (%d iterations vs previous of %d)" %\
                (ITERATIONS_TO_CONVERGE, Counter())
        return result

class TestWaterTutorial(ForceBalanceSystemTest):
    def setup_method(self, method):
        super(TestWaterTutorial, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(self.cwd, '..','..', 'studies','001_water_tutorial'))
        targets = tarfile.open('targets.tar.bz2','r')
        targets.extractall()
        targets.close()
        self.input_file='very_simple.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_WATER_RESULTS"
        self.expected_results = EXPECTED_WATER_RESULTS
        self.absolute_tolerance = 0.005

    def test_water_tutorial(self):
        """Check water tutorial study runs without errors"""
        self.run_optimizer()


class TestVoelzStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        super(TestVoelzStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '..', '..', 'studies', '009_voelz_nspe'))
        self.input_file='options.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)

    def test_voelz_study(self):
        """Check voelz study runs without errors"""
        self.run_optimizer(check_result=False, check_iter=False)

class TestBromineStudy(ForceBalanceSystemTest):

    def setup_method(self, method):
        super(TestBromineStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '..', '..', 'studies', '003_liquid_bromine'))
        self.input_file='optimize.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_BROMINE_RESULTS"
        self.expected_results = EXPECTED_BROMINE_RESULTS
        self.absolute_tolerance = 0.05

    def test_bromine_study(self):
        """Check liquid bromine study converges to expected results"""
        self.run_optimizer()

class TestThermoBromineStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        super(TestThermoBromineStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '../../studies/004_thermo_liquid_bromine'))
        self.input_file='optimize.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_BROMINE_RESULTS"
        self.expected_results = EXPECTED_BROMINE_RESULTS
        self.absolute_tolerance = 0.05

    def test_thermo_bromine_study(self):
        """Check liquid bromine study (Thermo target) converges to expected results"""
        self.run_optimizer()

class TestEvaluatorBromineStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        pytest.importorskip("evaluator")
        super(TestEvaluatorBromineStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '..', '..', 'studies', '003d_evaluator_liquid_bromine'))
        ## Extract targets archive.
        targets = tarfile.open('targets.tar.gz','r')
        targets.extractall()
        targets.close()
        ## Start the estimator server.
        import subprocess, time
        self.estimator_process = subprocess.Popen([
            "python", "run_server.py", "-ngpus=0", "-ncpus=1"
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        ## Give the server time to start.
        time.sleep(5)
        self.input_file='gradient.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)

    def teardown_method(self):
        self.estimator_process.terminate()
        shutil.rmtree("working_directory")
        shutil.rmtree("stored_data")
        super(TestEvaluatorBromineStudy, self).teardown_method()

    def test_bromine_study(self):
        """Check bromine study produces objective function and gradient in expected range """
        objective = self.get_objective()
        data      = objective.Full(np.zeros(objective.FF.np),1,verbose=True)
        X, G, H   = data['X'], data['G'], data['H']
        msgX="\nCalculated objective function is outside expected range.\n If this seems reasonable, update EXPECTED_EVALUATOR_BROMINE_OBJECTIVE in test_system.py with these values"
        np.testing.assert_allclose(EXPECTED_EVALUATOR_BROMINE_OBJECTIVE, X, atol=60, err_msg=msgX)
        msgG="\nCalculated gradient is outside expected range.\n If this seems reasonable, update EXPECTED_EVALUATOR_BROMINE_GRADIENT in test_system.py with these values"
        np.testing.assert_allclose(EXPECTED_EVALUATOR_BROMINE_GRADIENT, G, atol=2000, err_msg=msgG)

class TestLipidStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        super(TestLipidStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '../../studies/010_lipid_study'))
        self.input_file='simple.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_LIPID_RESULTS"
        self.expected_results = EXPECTED_LIPID_RESULTS
        self.absolute_tolerance = 0.100

    def test_lipid_study(self):
        """Check lipid tutorial study runs without errors"""
        self.run_optimizer()

class TestImplicitSolventHFEStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        super(TestImplicitSolventHFEStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '..', '..', 'studies', '012_implicit_solvent_hfe'))
        self.input_file='optimize.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_ETHANOL_RESULTS"
        self.expected_results = EXPECTED_ETHANOL_RESULTS
        self.absolute_tolerance = 0.020
 
    def test_implicit_solvent_hfe_study(self):
        """Check implicit hydration free energy study (Hydration target) converges to expected results"""
        self.run_optimizer(check_result=False, check_iter=False, use_pvals=True)

class TestOpenFFTorsionProfileStudy(ForceBalanceSystemTest):
    def setup_method(self, method):
        pytest.importorskip("openforcefield", minversion="0.4")
        pytest.importorskip("openeye.oechem")
        super(TestOpenFFTorsionProfileStudy, self).setup_method(method)
        cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(cwd, '..', '..', 'studies', '023_torsion_relaxed'))
        targets = tarfile.open('targets.tar.gz','r')
        targets.extractall()
        targets.close()
        self.input_file='optimize_minimal.in'
        self.logger.debug("\nSetting input file to '%s'\n" % self.input_file)
        self.expected_results_name = "EXPECTED_OPENFF_TORSIONPROFILE_RESULTS"
        self.expected_results = EXPECTED_OPENFF_TORSIONPROFILE_RESULTS
        self.absolute_tolerance = 0.001

    def test_openff_torsionprofile_study(self):
        """Check OpenFF torsion profile optimization converges to expected results"""
        self.run_optimizer(check_iter=False)



from builtins import object
import os, re
import forcebalance.output

forcebalance.output.getLogger("forcebalance.test").propagate=False

os.chdir(os.path.dirname(__file__))
__all__ = [module[:-3] for module in sorted(os.listdir('.'))
           if re.match("^test_.*\.py$",module)]


class ForceBalanceTestCase(object):
    @classmethod
    def setup_class(cls):
        """Override default test case constructor to set longMessage=True, reset cwd after test
        @override unittest.TestCase.__init(methodName='runTest')"""

        cls.logger = forcebalance.output.getLogger('forcebalance.test.' + __name__[5:])
        cls.start_directory = os.getcwd()

        # unset this env to prevent error in mdrun
        if 'OMP_NUM_THREADS' in os.environ:
            os.environ.pop('OMP_NUM_THREADS')

    def setup_method(self, method):
        pass

    def teardown_method(self):
        os.chdir(self.start_directory)



from __future__ import absolute_import
import os
import sys
import shutil
import pytest
from forcebalance.parser import parse_inputs
from forcebalance.forcefield import FF
from forcebalance.objective import Objective
from forcebalance.optimizer import Optimizer
from .__init__ import ForceBalanceTestCase

class TestWaterTutorial(ForceBalanceTestCase):
    def setup_method(self, method):
        super(TestWaterTutorial, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        # copy folder 'files/test_liquid' into a new folder 'files/test_liquid.run'
        os.chdir(os.path.join(self.cwd, 'files'))
        tmpfolder = os.path.join(self.cwd, 'files', 'test_liquid.run')
        source_folder = os.path.join(self.cwd, 'files', 'test_liquid')
        if os.path.isdir(tmpfolder):
            shutil.rmtree(tmpfolder)
        shutil.copytree(source_folder, tmpfolder)
        os.chdir(tmpfolder)

    def teardown_method(self):
        # remove temporary folder 'files/test_liquid.run'
        tmpfolder = os.path.join(self.cwd, 'files', 'test_liquid.run')
        if os.path.isdir(tmpfolder):
            shutil.rmtree(tmpfolder)
        super(TestWaterTutorial, self).teardown_method()

    def test_liquid(self):
        """Check liquid target with existing simulation data"""
        if sys.version_info <= (2,7):
            pytest.skip("Existing pickle file only works with Python 3")

        self.logger.debug("Setting input file to 'single.in'\n")
        input_file ='single.in'

        ## The general options and target options that come from parsing the input file
        self.logger.debug("Parsing inputs...\n")
        options, tgt_opts = parse_inputs(input_file)
        self.logger.debug("options:\n%s\n\ntgt_opts:\n%s\n\n" % (str(options), str(tgt_opts)))

        forcefield  = FF(options)
        objective   = Objective(options, tgt_opts, forcefield)
        ## The optimizer component of the project
        self.logger.debug("Creating optimizer: ")
        optimizer   = Optimizer(options, objective, forcefield)
        assert isinstance(optimizer, Optimizer), "Expected forcebalance optimizer object"
        self.logger.debug(str(optimizer))

        ## Actually run the optimizer.
        self.logger.debug("\nDone setting up! Running optimizer...")
        result = optimizer.Run()
        self.logger.debug("\nOptimizer finished. Final results:")
        self.logger.debug(str(result))

        liquid_obj_value = optimizer.Objective.ObjDict['Liquid']['x']
        assert liquid_obj_value < 20, "Liquid objective function should give < 20 (about 17.23) total value."




from __future__ import absolute_import
from builtins import range
from .__init__ import ForceBalanceTestCase
import numpy
import forcebalance
import re, sys
from math import cos, sin, pi
import pytest

class TestFiniteDifference(ForceBalanceTestCase):
    @classmethod
    def setup_class(cls):
        super(TestFiniteDifference, cls).setup_class()
        # functions is a list of 3-tuples containing (function p, d/dp, d^2/dp^2)
        cls.functions = []

        # f(x) = 2x
        cls.functions.append((lambda x: 2,
                              lambda x,p: 0,
                              lambda x,p: 0))
        # f(x) = cos(x)
        cls.functions.append((lambda x: cos(x[0]),
                              lambda x,p: -sin(x[0])*(p==0),
                              lambda x,p: -cos(x[0])*(p==0)))
        # f(x) = x^2
        cls.functions.append((lambda x: x[0]**2,
                              lambda x,p: 2*x[0]*(p==0),
                              lambda x,p: 2*(p==0)))

    def test_fdwrap(self):
        """Check fdwrap properly wraps function"""
        for func in self.functions:
            msg = "\nfdwrap alters function behavior"
            f = forcebalance.finite_difference.fdwrap(func[0], [0]*3, 0)
            self.logger.debug("Checking to make sure fdwrap returns a function\n")
            assert callable(f), "fdwrap did not return a function" #hasattr(f, '__call__'),  "fdwrap did not return a function"

            # some test values
            for x in range(-10, 11):
                assert abs(f(x)-func[0]([x, 0, 0])) < 1e-7

    def test_fd_stencils(self):
        """Check finite difference stencils return approximately correct results"""
        func = lambda x: x[0]**2
        fd_stencils = [function for function in dir(forcebalance.finite_difference) if re.match('^f..?d.p$', function)]

        self.logger.debug("Comparing fd stencils against some simple functions\n")
        for func in self.functions:
            for p in range(1):
                for x in range(10):
                    input = [0,0,0]
                    input[p]=x
                    f=forcebalance.finite_difference.fdwrap(func[0], input, p)
                    for stencil in fd_stencils:
                        fd = eval("forcebalance.finite_difference.%s" % stencil)
                        result = fd(f, .0001)
                        if re.match('^f..d.p$', stencil):
                            assert abs(result[0]-func[1](input,p)) < 1e-3
                            assert abs(result[1]-func[2](input,p)) < 1e-3
                        else:
                            assert abs(result-func[1](input,p)) < 1e-3



from __future__ import absolute_import
from builtins import str
from builtins import object
import os, re
import forcebalance
import numpy
import inspect
import pytest
from .__init__ import ForceBalanceTestCase

class TestImplemented(ForceBalanceTestCase):
    def test_implemented_targets_derived_from_target(self):
        """Check classes listed in Implemented_Targets are derived from Target"""
        for key in forcebalance.objective.Implemented_Targets.keys():
            self.logger.debug("Assert %s is subclass of target\n" % str(forcebalance.objective.Implemented_Targets[key]))
            assert issubclass(forcebalance.objective.Implemented_Targets[key],forcebalance.target.Target)

    def test_no_unlisted_classes_derived_from_Target(self):
        """Check for unknown omissions from Implemented_Targets

        Check to make sure any classes derived from Target are either
        listed in Implemented_Targets or in the exclusion list in this
        test case
        """
        forcebalance_modules=[module[:-3] for module in os.listdir(forcebalance.__path__[0])
                    if re.compile(".*\.py$").match(module)
                    and module not in ["__init__.py"]]
        for module in forcebalance_modules:
            # LPW: I don't think dcdlib should be imported this way.
            self.logger.debug(module)
            # Skip over smirnoff_hack because it is not intended to contain any Target implementations.
            if module in ["_dcdlib", "smirnoff_hack"]: continue
            m = __import__('forcebalance.' + module)
            objs = dir(eval('m.' + module))
            self.logger.debug(objs)
            for obj in objs:
                obj = eval('m.'+module+'.'+obj)
                if inspect.isclass(obj) and issubclass(obj, forcebalance.target.Target):
                    implemented = [i for i in forcebalance.objective.Implemented_Targets.values()]
                    # list of documented exceptions
                    # Basically, platform-independent targets are excluded.
                    exclude = ['Target',
                            'AbInitio',
                            'Interaction',
                            'Interaction_GMX',
                            'Liquid',
                            'Lipid',
                            'BindingEnergy',
                            'LeastSquares',
                            'Vibration',
                            'Thermo',
                            'Hydration',
                            'Moments', 
                            'OptGeoTarget',
                            'TorsionProfileTarget']
                    self.logger.debug(obj)
                    if obj not in implemented and obj.__name__ not in exclude:
                        pytest.fail("Unknown class '%s' not listed in Implemented_Targets" % obj.__name__)

class TestPenalty(ForceBalanceTestCase):
    def setup_method(self, method):
        super(TestPenalty, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(self.cwd, 'files'))
        self.options=forcebalance.parser.gen_opts_defaults.copy()
        self.options.update({
                'root': os.getcwd(),
                'penalty_additive': 0.01,
                'jobtype': 'NEWTON',
                'forcefield': ['cc-pvdz-overlap-original.gbs']})
        os.chdir(self.options['root'])

        self.ff = forcebalance.forcefield.FF(self.options)
        self.np = self.ff.np

        self.penalties = []
        for ptype in forcebalance.objective.Penalty.Pen_Names.keys():
            penalty = forcebalance.objective.Penalty(ptype,
                                self.ff,
                                self.options['penalty_additive'],
                                self.options['penalty_multiplicative'],
                                self.options['penalty_hyperbolic_b'],
                                self.options['penalty_alpha'])
            self.penalties.append(penalty)

    def test_penalty_compute(self):
        """Check penalty computation functions"""
        objective = {'G': numpy.zeros((9)),
         'H': numpy.diag((1,)*9),
         'X': 1}
        for penalty in self.penalties:
            result = penalty.compute([1]*self.np, objective)
            assert isinstance(result,tuple)
            # more tests go here

class ObjectiveTests(object):
    def test_target_zero_order_terms(self):
        """Check zero order target terms"""
        obj = self.objective.Target_Terms(numpy.array([.5]*self.ff.np), Order=0)
        assert isinstance(obj, dict)
        assert "X" in obj
        assert "G" in obj
        assert "H" in obj
        assert int(obj["X"]) != 0
        assert obj["G"].any() == False
        assert (obj["H"] == numpy.diag([1]*self.ff.np)).all()

    def test_target_first_order_terms(self):
        """Check first order target terms"""
        obj = self.objective.Target_Terms(numpy.array([.5]*self.ff.np), Order=1)
        assert isinstance(obj, dict)
        assert "X" in obj
        assert "G" in obj
        assert "H" in obj

    def test_target_second_order_terms(self):
        """Check second order target terms"""
        obj = self.objective.Target_Terms(numpy.array([.5]*self.ff.np), Order=2)
        assert isinstance(obj, dict)
        assert "X" in obj
        assert "G" in obj
        assert "H" in obj

    def test_indicate(self):
        """Check objective.indicate() runs without errors"""
        self.objective.Indicate()

class TestWaterObjective(ForceBalanceTestCase, ObjectiveTests):
    def setup_method(self, method):
        super(TestWaterObjective, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(self.cwd, 'files'))
        self.options=forcebalance.parser.gen_opts_defaults.copy()
        self.options.update({
                'root': os.getcwd(),
                'penalty_additive': 0.01,
                'jobtype': 'NEWTON',
                'forcefield': ['water.itp']})
        os.chdir(self.options['root'])

        self.logger.debug("\nUsing the following options:\n%s\n" % str(self.options))

        self.tgt_opts = [ forcebalance.parser.tgt_opts_defaults.copy() ]
        self.tgt_opts[0].update({"type" : "ABINITIO_GMX", "name" : "cluster-06"})
        self.ff = forcebalance.forcefield.FF(self.options)

        self.objective = forcebalance.objective.Objective(self.options, self.tgt_opts,self.ff)

class TestBromineObjective(ForceBalanceTestCase, ObjectiveTests):
    def setup_method(self, method):
        super(TestBromineObjective, self).setup_method(method)
        self.cwd = os.path.dirname(os.path.realpath(__file__))
        os.chdir(os.path.join(self.cwd, 'files'))
        self.options=forcebalance.parser.gen_opts_defaults.copy()
        self.options.update({
                'root': os.getcwd(),
                'penalty_additive': 0.01,
                'jobtype': 'NEWTON',
                'forcefield': ['bro.itp']})
        os.chdir(self.options['root'])

        self.logger.debug("\nUsing the following options:\n%s\n" % str(self.options))

        self.tgt_opts = [ forcebalance.parser.tgt_opts_defaults.copy() ]
        self.tgt_opts[0].update({"type" : "LIQUID_GMX", "name" : "LiquidBromine"})
        self.ff = forcebalance.forcefield.FF(self.options)

        self.objective = forcebalance.objective.Objective(self.options, self.tgt_opts,self.ff)



from __future__ import absolute_import
import forcebalance
import shutil
from .test_target import TargetTests # general targets tests defined in test_target.py
import pytest
"""
The testing functions for this class are located in test_target.py.
"""
class TestLiquid_OpenMM(TargetTests):
    def setup_method(self, method):
        pytest.skip("Needs optimizing to reduce runtime")
        super(TestLiquid_OpenMM, self).setup_method(method)
        # settings specific to this target
        self.options.update({
                'jobtype': 'NEWTON',
                'forcefield': ['dms.xml']})

        self.tgt_opt.update({'type':'LIQUID_OPENMM',
            'name':'dms-liquid'})

        self.ff = forcebalance.forcefield.FF(self.options)

        self.ffname = self.options['forcefield'][0][:-3]
        self.filetype = self.options['forcefield'][0][-3:]
        self.mvals = [.5]*self.ff.np

        self.target = forcebalance.openmmio.Liquid_OpenMM(self.options, self.tgt_opt, self.ff)
        self.target.stage(self.mvals)

    def teardown_method(self):
        shutil.rmtree('temp')
        super(TestLiquid_OpenMM, self).teardown_method()

class TestInteraction_OpenMM(TargetTests):

    def setup_method(self, method):
        super(TestInteraction_OpenMM, self).setup_method(method)
        # TargetTests.setup_class(cls)
        # settings specific to this target
        self.options.update({
                'jobtype': 'NEWTON',
                'forcefield': ['dms.xml']})

        self.tgt_opt.update({"type" : "Interaction_OpenMM",
                            "name" : "S2EPose",
                            "fragment1" : "1-9",
                            "fragment2" : "10-18"})

        self.ff = forcebalance.forcefield.FF(self.options)

        self.ffname = self.options['forcefield'][0][:-3]
        self.filetype = self.options['forcefield'][0][-3:]
        self.mvals = [.5]*self.ff.np

        self.target = forcebalance.openmmio.Interaction_OpenMM(self.options, self.tgt_opt, self.ff)

    def teardown_method(self):
        shutil.rmtree('temp')
        super(TestInteraction_OpenMM, self).teardown_method()




""" @package forcebalance.smirnoff SMIRNOFF force field support.

@author Lee-Ping Wang
@date 12/2018
"""
from __future__ import division

from builtins import zip
from builtins import range
import os
from forcebalance import BaseReader
from forcebalance.abinitio import AbInitio
from forcebalance.binding import BindingEnergy
from forcebalance.liquid import Liquid
from forcebalance.interaction import Interaction
from forcebalance.moments import Moments
from forcebalance.hydration import Hydration
from forcebalance.vibration import Vibration
from forcebalance.opt_geo_target import OptGeoTarget
from forcebalance.torsion_profile import TorsionProfileTarget
import networkx as nx
import numpy as np
import sys
from forcebalance.finite_difference import *
import pickle
import shutil
from copy import deepcopy
from forcebalance.engine import Engine
from forcebalance.molecule import *
from forcebalance.chemistry import *
from forcebalance.nifty import *
from forcebalance.nifty import _exec
from collections import OrderedDict, defaultdict, Counter
from forcebalance.output import getLogger
from forcebalance.openmmio import OpenMM, UpdateSimulationParameters
import json

logger = getLogger(__name__)
try:
    from simtk.openmm.app import *
    from simtk.openmm import *
    from simtk.unit import *
    import simtk.openmm._openmm as _openmm
except:
    pass

try:
    # import the hack for openforcefield to improve performance by 10x
    from forcebalance import smirnoff_hack
    # Import the SMIRNOFF forcefield engine and some useful tools
    from openforcefield.typing.engines.smirnoff import ForceField as OpenFF_ForceField
    # QYD: name of class are modified to avoid colliding with ForceBalance Molecule
    from openforcefield.topology import Molecule as OffMolecule
    from openforcefield.topology import Topology as OffTopology
except:
    pass

## pdict is a useless variable if the force field is XML.
pdict = "XML_Override"

def smirnoff_analyze_parameter_coverage(forcefield, tgt_opts):
    printcool("SMIRNOFF Parameter Coverage Analysis")
    assert hasattr(forcefield, 'offxml'), "Only SMIRNOFF Force Field is supported"
    parameter_assignment_data = defaultdict(list)
    parameter_counter = Counter()
    # The openforcefield.typing.engines.smirnoff.ForceField object should now be contained in forcebalance.forcefield.FF
    ff = forcefield.openff_forcefield
    # analyze each target
    for tgt_option in tgt_opts:
        target_path = os.path.join('targets', tgt_option['name'])
        # aggregate mol2 file paths from all targets
        mol2_paths = []
        if tgt_option['type'] == 'OPTGEOTARGET_SMIRNOFF':
            # parse optgeo_options_txt and get the names of the mol2 files
            optgeo_options_txt = os.path.join(target_path, tgt_option['optgeo_options_txt'])
            sys_opts = forcebalance.opt_geo_target.OptGeoTarget.parse_optgeo_options(optgeo_options_txt)
            mol2_paths = [os.path.join(target_path,fnm) for sysopt in sys_opts.values() for fnm in sysopt['mol2']]
        elif tgt_option['type'].endswith('_SMIRNOFF'):
            mol2_paths = [os.path.join(target_path,fnm) for fnm in tgt_option['mol2']]
        # analyze SMIRKs terms
        for mol_fnm in mol2_paths:
            # we work with one file at a time to avoid the topology sliently combine "same" molecules
            openff_mol = OffMolecule.from_file(mol_fnm)
            off_topology = OffTopology.from_molecules([openff_mol])
            molecule_force_list = ff.label_molecules(off_topology)
            for mol_idx, mol_forces in enumerate(molecule_force_list):
                for force_tag, force_dict in mol_forces.items():
                    # e.g. force_tag = 'Bonds'
                    for atom_indices, parameter in force_dict.items():
                        param_dict = {'id': parameter.id, 'smirks': parameter.smirks, 'type':force_tag, 'atoms': list(atom_indices),}
                        parameter_assignment_data[mol_fnm].append(param_dict)
                        parameter_counter[parameter.smirks] += 1
    # write out parameter assignment data
    out_json_path = os.path.join(forcefield.root, 'smirnoff_parameter_assignments.json')
    with open(out_json_path, 'w') as jsonfile:
        json.dump(parameter_assignment_data, jsonfile, indent=2)
        logger.info("Force field assignment data written to %s\n" % out_json_path)
    # print parameter coverages
    logger.info("%4s %-100s   %10s\n" % ("idx", "Parameter", "Count"))
    logger.info("-"*118 + '\n')
    n_covered = 0
    for i,p in enumerate(forcefield.plist):
        smirks = p.split('/')[-1]
        logger.info('%4i %-100s : %10d\n' % (i, p, parameter_counter[smirks]))
        if parameter_counter[smirks] > 0:
            n_covered += 1
    logger.info("SNIRNOFF Parameter Coverage Analysis result: %d/%d parameters are covered.\n" % (n_covered, len(forcefield.plist)))
    logger.info("-"*118 + '\n')

class SMIRNOFF_Reader(BaseReader):
    """ Class for parsing OpenMM force field files. """
    def __init__(self,fnm):
        ## Initialize the superclass. :)
        super(SMIRNOFF_Reader,self).__init__(fnm)
        ## The parameter dictionary (defined in this file)
        self.pdict  = pdict

    def build_pid(self, element, parameter):
        """ Build the parameter identifier (see _link_ for an example)
        @todo Add a link here """
        ParentType = ".".join([i.tag for i in list(element.iterancestors())][::-1][1:])
        InteractionType = element.tag
        try:
            Involved = element.attrib["smirks"]
            return "/".join([ParentType, InteractionType, parameter, Involved])
        except:
            logger.info("Minor warning: Parameter ID %s doesn't contain any SMIRKS patterns, redundancies are possible\n" % ("/".join([InteractionType, parameter])))
            return "/".join([ParentType, InteractionType, parameter])

def assign_openff_parameter(ff, new_value, pid):
    """
    Assign a SMIRNOFF parameter given the openforcefield.ForceField object, the desired parameter value,
    and the parameter's unique ID.
    """
    # Split the parameter's unique ID into four fields using a slash:
    # Input: ProperTorsions/Proper/k1/[*:1]~[#6X3:2]:[#6X3:3]~[*:4]
    # Output: ProperTorsions, Proper, k1, [*:1]~[#6X3:2]:[#6X3:3]~[*:4]
    # The first, third and fourth fields will be used for parameter assignment.
    # We use "value_name" to describe names of individual numerical values within a single parameter type
    # e.g. k1 in the above example.

    # QYD: cache the parameter finding procedure, then directly change the _value of the quantity
    # Note: This cache requires the quantity does not get overwritten, which is True since this function is the only
    # place we modify the OpenFF ForceField parameters.
    if not hasattr(ff, '_forcebalance_assign_parameter_map'):
        ff._forcebalance_assign_parameter_map = dict()
    if pid not in ff._forcebalance_assign_parameter_map:
        (handler_name, tag_name, value_name, smirks) = pid.split('/')
        # Get the OpenFF parameter object
        parameter = ff.get_parameter_handler(handler_name).parameters[smirks]
        if hasattr(parameter, value_name):
            # If the value name is an attribute of the parameter then we set it directly.
            unit = getattr(parameter, value_name).unit
            # Get the quantity of the parameter in the OpenFF forcefield object
            param_quantity = getattr(parameter, value_name)
        elif value_name in parameter._cosmetic_attribs:
            param_quantity = None
        else:
            # If the value name is a periodic attribute (say, k1) then we need to use
            # a regex to split the value name into 'k' and '1', then set the appropriate
            # value in the k-list
            attribute_split = re.split(r'(\d+)', value_name)
            # print(attribute_split)
            # assert len(attribute_split) == 2
            assert hasattr(parameter, attribute_split[0]), "%s.%s not exist" % (parameter, attribute_split[0])
            # attribute_split[0] is a string such as 'k'
            value_name = attribute_split[0]
            # parameter_index is the position of k1 in the values associated with 'k'
            parameter_index = int(attribute_split[1]) - 1
            # Get the list of values, update the appropriate one and then set the new attribute to the updated list
            value_list = getattr(parameter, value_name)
            # Get the quantity of the parameter in the OpenFF forcefield object
            param_quantity = value_list[parameter_index]
        # save the found quantity in cache
        ff._forcebalance_assign_parameter_map[pid] = param_quantity
    else:
        param_quantity = ff._forcebalance_assign_parameter_map[pid]
    # set new_value directly in the quantity
    if param_quantity is not None:
        param_quantity._value = new_value

class SMIRNOFF(OpenMM):

    """ Derived from Engine object for carrying out OpenMM calculations that use the SMIRNOFF force field. """

    def __init__(self, name="openmm", **kwargs):
        self.valkwd = ['ffxml', 'pdb', 'mol2', 'platname', 'precision', 'mmopts', 'vsite_bonds', 'implicit_solvent', 'restrain_k', 'freeze_atoms']
        super(SMIRNOFF,self).__init__(name=name, **kwargs)

    def readsrc(self, **kwargs):
        """
        SMIRNOFF simulations always require the following passed in via kwargs:

        Parameters
        ----------
        pdb : string
            Name of a .pdb file containing the topology of the system
        mol2 : list
            A list of .mol2 file names containing the molecule/residue templates of the system

        Also provide 1 of the following, containing the coordinates to be used:
        mol : Molecule
            forcebalance.Molecule object
        coords : string
            Name of a file (readable by forcebalance.Molecule)
            This could be the same as the pdb argument from above.
        """

        pdbfnm = kwargs.get('pdb')
        # Determine the PDB file name.
        if not pdbfnm:
            raise RuntimeError('Name of PDB file not provided.')
        elif not os.path.exists(pdbfnm):
            logger.error("%s specified but doesn't exist\n" % pdbfnm)
            raise RuntimeError

        if 'mol' in kwargs:
            self.mol = kwargs['mol']
        elif 'coords' in kwargs:
            if not os.path.exists(kwargs['coords']):
                logger.error("%s specified but doesn't exist\n" % kwargs['coords'])
                raise RuntimeError
            self.mol = Molecule(kwargs['coords'])
        else:
            logger.error('Must provide either a molecule object or coordinate file.\n')
            raise RuntimeError

        # Here we cannot distinguish the .mol2 files linked by the target
        # vs. the .mol2 files to be provided by the force field.
        # But we can assume that these files should exist when this function is called.

        self.mol2 = kwargs.get('mol2')
        if self.mol2:
            for fnm in self.mol2:
                if not os.path.exists(fnm):
                    if hasattr(self, 'FF') and fnm in self.FF.fnms: continue
                    logger.error("%s doesn't exist" % fnm)
                    raise RuntimeError
        else:
            logger.error("Must provide a list of .mol2 files.\n")

        self.abspdb = os.path.abspath(pdbfnm)
        mpdb = Molecule(pdbfnm)
        for i in ["chain", "atomname", "resid", "resname", "elem"]:
            self.mol.Data[i] = mpdb.Data[i]

        # Store a separate copy of the molecule for reference restraint positions.
        self.ref_mol = deepcopy(self.mol)

    def prepare(self, pbc=False, mmopts={}, **kwargs):

        """
        Prepare the calculation.  Note that we don't create the
        Simulation object yet, because that may depend on MD
        integrator parameters, thermostat, barostat etc.

        This is mostly copied and modified from openmmio.py's OpenMM.prepare(),
        but we are calling ForceField() from the OpenFF toolkit and ignoring
        AMOEBA stuff.
        """
        self.pdb = PDBFile(self.abspdb)

        # Create the OpenFF ForceField object.
        if hasattr(self, 'FF'):
            self.offxml = [self.FF.offxml]
            self.forcefield = self.FF.openff_forcefield
        else:
            self.offxml = listfiles(kwargs.get('offxml'), 'offxml', err=True)
            self.forcefield = OpenFF_ForceField(*self.offxml)

        ## Load mol2 files for smirnoff topology
        openff_mols = []
        for fnm in self.mol2:
            try:
                mol = OffMolecule.from_file(fnm)
            except Exception as e:
                logger.error("Error when loading %s" % fnm)
                raise e
            openff_mols.append(mol)
        self.off_topology = OffTopology.from_openmm(self.pdb.topology, unique_molecules=openff_mols)

        # used in create_simulation()
        self.mod = Modeller(self.pdb.topology, self.pdb.positions)

        ## OpenMM options for setting up the System.
        self.mmopts = dict(mmopts)

        ## Specify frozen atoms and restraint force constant
        if 'restrain_k' in kwargs:
            self.restrain_k = kwargs['restrain_k']
        if 'freeze_atoms' in kwargs:
            self.freeze_atoms = kwargs['freeze_atoms'][:]

        ## Set system options from ForceBalance force field options.
        fftmp = False
        if hasattr(self,'FF'):
            self.mmopts['rigidWater'] = self.FF.rigid_water
            if not all([os.path.exists(f) for f in self.FF.fnms]):
                # If the parameter files don't already exist, create them for the purpose of
                # preparing the engine, but then delete them afterward.
                fftmp = True
                self.FF.make(np.zeros(self.FF.np))

        ## Set system options from periodic boundary conditions.
        self.pbc = pbc
        ## print warning for 'nonbonded_cutoff' keywords
        if 'nonbonded_cutoff' in kwargs:
            logger.warning("nonbonded_cutoff keyword ignored because it's set in the offxml file\n")

        ## Generate OpenMM-compatible positions
        self.xyz_omms = []
        for I in range(len(self.mol)):
            position = self.mol.xyzs[I] * angstrom
            # xyz_omm = [Vec3(i[0],i[1],i[2]) for i in xyz]*angstrom
            # An extra step with adding virtual particles
            # mod = Modeller(self.pdb.topology, xyz_omm)
            # LPW commenting out because we don't have virtual sites yet.
            # mod.addExtraParticles(self.forcefield)
            if self.pbc:
                # Obtain the periodic box
                if self.mol.boxes[I].alpha != 90.0 or self.mol.boxes[I].beta != 90.0 or self.mol.boxes[I].gamma != 90.0:
                    logger.error('OpenMM cannot handle nonorthogonal boxes.\n')
                    raise RuntimeError
                box_omm = np.diag([self.mol.boxes[I].a, self.mol.boxes[I].b, self.mol.boxes[I].c]) * angstrom
            else:
                box_omm = None
            # Finally append it to list.
            self.xyz_omms.append((position, box_omm))

        ## Build a topology and atom lists.
        Top = self.pdb.topology
        Atoms = list(Top.atoms())
        Bonds = [(a.index, b.index) for a, b in list(Top.bonds())]

        # vss = [(i, [system.getVirtualSite(i).getParticle(j) for j in range(system.getVirtualSite(i).getNumParticles())]) \
        #            for i in range(system.getNumParticles()) if system.isVirtualSite(i)]
        self.AtomLists = defaultdict(list)
        self.AtomLists['Mass'] = [a.element.mass.value_in_unit(dalton) if a.element is not None else 0 for a in Atoms]
        self.AtomLists['ParticleType'] = ['A' if m >= 1.0 else 'D' for m in self.AtomLists['Mass']]
        self.AtomLists['ResidueNumber'] = [a.residue.index for a in Atoms]
        self.AtomMask = [a == 'A' for a in self.AtomLists['ParticleType']]
        self.realAtomIdxs = [i for i, a in enumerate(self.AtomMask) if a is True]
        if hasattr(self,'FF') and fftmp:
            for f in self.FF.fnms:
                os.unlink(f)

    def update_simulation(self, **kwargs):

        """
        Create the simulation object, or update the force field
        parameters in the existing simulation object.  This should be
        run when we write a new force field XML file.
        """
        if len(kwargs) > 0:
            self.simkwargs = kwargs

        # Because self.forcefield is being updated in forcebalance.forcefield.FF.make()
        # there is no longer a need to create a new force field object here.
        try:
            self.system = self.forcefield.create_openmm_system(self.off_topology)
        except Exception as error:
            logger.error("Error when creating system for %s" % self.mol2)
            raise error
        # Commenting out all virtual site stuff for now.
        # self.vsinfo = PrepareVirtualSites(self.system)
        self.nbcharges = np.zeros(self.system.getNumParticles())

        #----
        # If the virtual site parameters have changed,
        # the simulation object must be remade.
        #----
        # vsprm = GetVirtualSiteParameters(self.system)
        # if hasattr(self,'vsprm') and len(self.vsprm) > 0 and np.max(np.abs(vsprm - self.vsprm)) != 0.0:
        #     if hasattr(self, 'simulation'):
        #         delattr(self, 'simulation')
        # self.vsprm = vsprm.copy()

        if hasattr(self, 'simulation'):
            UpdateSimulationParameters(self.system, self.simulation)
        else:
            self.create_simulation(**self.simkwargs)

    def optimize(self, shot=0, align=True, crit=1e-4):
        return super(SMIRNOFF,self).optimize(shot=shot, align=align, crit=crit, disable_vsite=True)

    def interaction_energy(self, fraga, fragb):

        """
        Calculate the interaction energy for two fragments.
        Because this creates two new objects and requires passing in the mol2 argument,
        the codes are copied and modified from the OpenMM class.
        """

        self.update_simulation()

        if self.name == 'A' or self.name == 'B':
            logger.error("Don't name the engine A or B!\n")
            raise RuntimeError

        # Create two subengines.
        if hasattr(self,'target'):
            if not hasattr(self,'A'):
                self.A = SMIRNOFF(name="A", mol=self.mol.atom_select(fraga), mol2=self.mol2, target=self.target)
            if not hasattr(self,'B'):
                self.B = SMIRNOFF(name="B", mol=self.mol.atom_select(fragb), mol2=self.mol2, target=self.target)
        else:
            if not hasattr(self,'A'):
                self.A = SMIRNOFF(name="A", mol=self.mol.atom_select(fraga), mol2=self.mol2, platname=self.platname, \
                                  precision=self.precision, offxml=self.offxml, mmopts=self.mmopts)
            if not hasattr(self,'B'):
                self.B = SMIRNOFF(name="B", mol=self.mol.atom_select(fragb), mol2=self.mol2, platname=self.platname, \
                                  precision=self.precision, offxml=self.offxml, mmopts=self.mmopts)

        # Interaction energy needs to be in kcal/mol.
        D = self.energy()
        A = self.A.energy()
        B = self.B.energy()

        return (D - A - B) / 4.184

    def get_smirks_counter(self):
        """Get a counter for the time of appreance of each SMIRKS"""
        smirks_counter = Counter()
        molecule_force_list = self.forcefield.label_molecules(self.off_topology)
        for mol_idx, mol_forces in enumerate(molecule_force_list):
            for force_tag, force_dict in mol_forces.items():
                # e.g. force_tag = 'Bonds'
                for parameter in force_dict.values():
                    smirks_counter[parameter.smirks] += 1
        return smirks_counter

class Liquid_SMIRNOFF(Liquid):
    """ Condensed phase property matching using OpenMM. """
    def __init__(self,options,tgt_opts,forcefield):
        # Time interval (in ps) for writing coordinates
        self.set_option(tgt_opts,'force_cuda',forceprint=True)
        # Enable multiple timestep integrator
        self.set_option(tgt_opts,'mts_integrator',forceprint=True)
        # Enable ring polymer MD
        self.set_option(options,'rpmd_beads',forceprint=True)
        # List of .mol2 files for SMIRNOFF to set up the system
        self.set_option(tgt_opts,'mol2',forceprint=True)
        # OpenMM precision
        self.set_option(tgt_opts,'openmm_precision','precision',default="mixed")
        # OpenMM platform
        self.set_option(tgt_opts,'openmm_platform','platname',default="CUDA")
        # Name of the liquid coordinate file.
        self.set_option(tgt_opts,'liquid_coords',default='liquid.pdb',forceprint=True)
        # Name of the gas coordinate file.
        self.set_option(tgt_opts,'gas_coords',default='gas.pdb',forceprint=True)
        # Name of the surface tension coordinate file. (e.g. an elongated box with a film of water)
        self.set_option(tgt_opts,'nvt_coords',default='surf.pdb',forceprint=True)
        # Set the number of steps between MC barostat adjustments.
        self.set_option(tgt_opts,'mc_nbarostat')
        # Class for creating engine object.
        self.engine_ = SMIRNOFF
        # Name of the engine to pass to npt.py.
        self.engname = "smirnoff"
        # Command prefix.
        self.nptpfx = "bash runcuda.sh"
        if tgt_opts['remote_backup']:
            self.nptpfx += " -b"
        # Extra files to be linked into the temp-directory.
        self.nptfiles = []
        self.nvtfiles = []
        # Set some options for the polarization correction calculation.
        self.gas_engine_args = {}
        # Scripts to be copied from the ForceBalance installation directory.
        self.scripts = ['runcuda.sh']
        # Initialize the base class.
        super(Liquid_SMIRNOFF,self).__init__(options,tgt_opts,forcefield)
        # Send back the trajectory file.
        if self.save_traj > 0:
            self.extra_output = ['liquid-md.pdb', 'liquid-md.dcd']
        # These functions need to be called after self.nptfiles is populated
        self.post_init(options)

class AbInitio_SMIRNOFF(AbInitio):
    """ Force and energy matching using OpenMM. """
    def __init__(self,options,tgt_opts,forcefield):
        ## Default file names for coordinates and key file.
        self.set_option(tgt_opts,'pdb',default="conf.pdb")
        # List of .mol2 files for SMIRNOFF to set up the system
        self.set_option(tgt_opts,'mol2',forceprint=True)
        self.set_option(tgt_opts,'coords',default="all.gro")
        self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
        self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
        self.engine_ = SMIRNOFF
        ## Initialize base class.
        super(AbInitio_SMIRNOFF,self).__init__(options,tgt_opts,forcefield)

    def submit_jobs(self, mvals, AGrad=False, AHess=False):
        # we update the self.pgrads here so it's not overwritten in rtarget.py
        self.smirnoff_update_pgrads()

    def smirnoff_update_pgrads(self):
        """
        Update self.pgrads based on smirks present in mol2 files

        This can greatly improve gradients evaluation in big optimizations

        Note
        ----
        1. This function assumes the names of the forcefield parameters has the smirks as the last item
        2. This function assumes params only affect the smirks of its own. This might not be true if parameter_eval is used.
        """
        orig_pgrad_set = set(self.pgrad)
        # smirks to param_idxs map
        smirks_params_map = defaultdict(list)
        # New code for mapping smirks to mathematical parameter IDs
        for pname in self.FF.pTree:
            smirks = pname.rsplit('/',maxsplit=1)[-1]
            for pidx in self.FF.get_mathid(pname):
                smirks_params_map[smirks].append(pidx)
        pgrads_set = set()
        # get the smirks for this target, keep only the pidx corresponding to these smirks
        smirks_counter = self.engine.get_smirks_counter()
        for smirks in smirks_counter:
            if smirks_counter[smirks] > 0:
                pidx_list = smirks_params_map[smirks]
                # update the set of parameters present in this target
                pgrads_set.update(pidx_list)
        # this ensure we do not add any new items into self.pgrad
        pgrads_set.intersection_update(orig_pgrad_set)
        self.pgrad = sorted(list(pgrads_set))


class Vibration_SMIRNOFF(Vibration):
    """ Vibrational frequency matching using TINKER. """
    def __init__(self,options,tgt_opts,forcefield):
        ## Default file names for coordinates and key file.
        self.set_option(tgt_opts,'coords',default="input.pdb")
        self.set_option(tgt_opts,'pdb',default="conf.pdb")
        self.set_option(tgt_opts,'mol2',forceprint=True)
        self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
        self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
        self.engine_ = SMIRNOFF
        ## Initialize base class.
        super(Vibration_SMIRNOFF,self).__init__(options,tgt_opts,forcefield)

    def submit_jobs(self, mvals, AGrad=False, AHess=False):
        # we update the self.pgrads here so it's not overwritten in rtarget.py
        self.smirnoff_update_pgrads()

    def smirnoff_update_pgrads(self):
        """
        Update self.pgrads based on smirks present in mol2 files

        This can greatly improve gradients evaluation in big optimizations

        Note
        ----
        1. This function assumes the names of the forcefield parameters has the smirks as the last item
        2. This function assumes params only affect the smirks of its own. This might not be true if parameter_eval is used.
        """
        orig_pgrad_set = set(self.pgrad)
        # smirks to param_idxs map
        smirks_params_map = defaultdict(list)
        for pname in self.FF.pTree:
            smirks = pname.rsplit('/',maxsplit=1)[-1]
            for pidx in self.FF.get_mathid(pname):
                smirks_params_map[smirks].append(pidx)
        pgrads_set = set()
        # get the smirks for this target, keep only the pidx corresponding to these smirks
        smirks_counter = self.engine.get_smirks_counter()
        for smirks in smirks_counter:
            if smirks_counter[smirks] > 0:
                pidx_list = smirks_params_map[smirks]
                # update the set of parameters present in this target
                pgrads_set.update(pidx_list)
        # this ensure we do not add any new items into self.pgrad
        pgrads_set.intersection_update(orig_pgrad_set)
        self.pgrad = sorted(list(pgrads_set))


class OptGeoTarget_SMIRNOFF(OptGeoTarget):
    """ Optimized geometry fitting using SMIRNOFF format powered by OpenMM """
    def __init__(self,options,tgt_opts,forcefield):
        self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
        self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
        self.engine_ = SMIRNOFF
        ## Initialize base class.
        super(OptGeoTarget_SMIRNOFF,self).__init__(options,tgt_opts,forcefield)

    def create_engines(self, engine_args):
        """ create a dictionary of self.engines = {sysname: Engine} """
        self.engines = OrderedDict()
        for sysname, sysopt in self.sys_opts.items():
            # SMIRNOFF is a subclass of OpenMM engine but it requires the mol2 input
            # note: OpenMM.mol is a Molecule class instance;  mol2 is a file format.
            # path to .pdb file
            pdbpath = os.path.join(self.root, self.tgtdir, sysopt['topology'])
            # a list of paths to .mol2 files
            mol2path = [os.path.join(self.root, self.tgtdir, f) for f in sysopt['mol2']]
            # use the PDB file with topology
            M = Molecule(os.path.join(self.root, self.tgtdir, sysopt['topology']))
            # replace geometry with values from xyz file for higher presision
            M0 = Molecule(os.path.join(self.root, self.tgtdir, sysopt['geometry']))
            M.xyzs = M0.xyzs
            # here mol=M is given for the purpose of using the topology from the input pdb file
            # if we don't do this, pdb=top.pdb option will only copy some basic information but not the topology into OpenMM.mol (openmmio.py line 615)
            self.engines[sysname] = self.engine_(target=self, mol=M, name=sysname, pdb=pdbpath, mol2=mol2path, **engine_args)
        self.build_system_mval_masks()

    def build_system_mval_masks(self):
        """
        Build a mask of mvals for each system, to speed up finite difference gradients

        Note
        ----
        1. This function assumes the names of the forcefield parameters has the smirks as the last item
        2. This function assumes params only affect the smirks of its own. This might not be true if parameter_eval is used.
        """
        # only need to build once
        if hasattr(self, 'system_mval_masks'): return
        n_params = len(self.FF.map)
        # default mask with all False
        system_mval_masks = {sysname: np.zeros(n_params, dtype=bool) for sysname in self.sys_opts}
        orig_pgrad_set = set(self.pgrad)
        # smirks to param_idxs map
        smirks_params_map = defaultdict(list)
        # New code for mapping smirks to mathematical parameter IDs
        for pname in self.FF.pTree:
            smirks = pname.rsplit('/',maxsplit=1)[-1]
            # print("pname %s mathid %s -> smirks %s" % (pname, str(self.FF.get_mathid(pname)), smirks))
            for pidx in self.FF.get_mathid(pname):
                smirks_params_map[smirks].append(pidx)
        # Old code for mapping smirks to mathematical parameter IDs
        # for pname, pidx in self.FF.map.items():
        #     smirks = pname.rsplit('/',maxsplit=1)[-1]
        #     smirks_params_map[smirks].append(pidx)
        # go over all smirks for each system
        for sysname in self.sys_opts:
            engine = self.engines[sysname]
            smirks_counter = engine.get_smirks_counter()
            for smirks in smirks_counter:
                if smirks_counter[smirks] > 0:
                    pidx_list = smirks_params_map[smirks]
                    # set mask value to True for present smirks
                    system_mval_masks[sysname][pidx_list] = True
        # finish
        logger.info("system_mval_masks is built for faster gradient evaluations")
        self.system_mval_masks = system_mval_masks

class TorsionProfileTarget_SMIRNOFF(TorsionProfileTarget):
    """ Force and energy matching using SMIRKS native Open Force Field (SMIRNOFF). """
    def __init__(self,options,tgt_opts,forcefield):
        ## Default file names for coordinates and key file.
        self.set_option(tgt_opts,'pdb',default="conf.pdb")
        # List of .mol2 files for SMIRNOFF to set up the system
        self.set_option(tgt_opts,'mol2',forceprint=True)
        self.set_option(tgt_opts,'coords',default="scan.xyz")
        self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
        self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
        self.engine_ = SMIRNOFF
        ## Initialize base class.
        super(TorsionProfileTarget_SMIRNOFF,self).__init__(options,tgt_opts,forcefield)

    def submit_jobs(self, mvals, AGrad=False, AHess=False):
        # we update the self.pgrads here so it's not overwritten in rtarget.py
        self.smirnoff_update_pgrads()

    def smirnoff_update_pgrads(self):
        """
        Update self.pgrads based on smirks present in mol2 files

        This can greatly improve gradients evaluation in big optimizations

        Note
        ----
        1. This function assumes the names of the forcefield parameters has the smirks as the last item
        2. This function assumes params only affect the smirks of its own. This might not be true if parameter_eval is used.
        """
        orig_pgrad_set = set(self.pgrad)
        # smirks to param_idxs map
        smirks_params_map = defaultdict(list)
        # New code for mapping smirks to mathematical parameter IDs
        for pname in self.FF.pTree:
            smirks = pname.rsplit('/',maxsplit=1)[-1]
            for pidx in self.FF.get_mathid(pname):
                smirks_params_map[smirks].append(pidx)
        pgrads_set = set()
        # get the smirks for this target, keep only the pidx corresponding to these smirks
        smirks_counter = self.engine.get_smirks_counter()
        for smirks in smirks_counter:
            if smirks_counter[smirks] > 0:
                pidx_list = smirks_params_map[smirks]
                # update the set of parameters present in this target
                pgrads_set.update(pidx_list)
        # this ensure we do not add any new items into self.pgrad
        pgrads_set.intersection_update(orig_pgrad_set)
        self.pgrad = sorted(list(pgrads_set))

# class BindingEnergy_SMIRNOFF(BindingEnergy):
#     """ Binding energy matching using OpenMM. """

#     def __init__(self,options,tgt_opts,forcefield):
#         self.engine_ = OpenMM
#         self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
#         self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
#         ## Initialize base class.
#         super(BindingEnergy_OpenMM,self).__init__(options,tgt_opts,forcefield)

# class Interaction_SMIRNOFF(Interaction):
#     """ Interaction matching using OpenMM. """
#     def __init__(self,options,tgt_opts,forcefield):
#         ## Default file names for coordinates and key file.
#         self.set_option(tgt_opts,'coords',default="all.pdb")
#         self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
#         self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
#         self.engine_ = OpenMM
#         ## Initialize base class.
#         super(Interaction_OpenMM,self).__init__(options,tgt_opts,forcefield)

# class Moments_SMIRNOFF(Moments):
#     """ Multipole moment matching using OpenMM. """
#     def __init__(self,options,tgt_opts,forcefield):
#         ## Default file names for coordinates and key file.
#         self.set_option(tgt_opts,'coords',default="input.pdb")
#         self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
#         self.set_option(tgt_opts,'openmm_platform','platname',default="Reference", forceprint=True)
#         self.engine_ = OpenMM
#         ## Initialize base class.
#         super(Moments_OpenMM,self).__init__(options,tgt_opts,forcefield)

# class Hydration_SMIRNOFF(Hydration):
#     """ Single point hydration free energies using OpenMM. """

#     def __init__(self,options,tgt_opts,forcefield):
#         ## Default file names for coordinates and key file.
#         # self.set_option(tgt_opts,'coords',default="input.pdb")
#         self.set_option(tgt_opts,'openmm_precision','precision',default="double", forceprint=True)
#         self.set_option(tgt_opts,'openmm_platform','platname',default="CUDA", forceprint=True)
#         self.engine_ = SMIRNOFF
#         self.engname = "smirnoff"
#         ## Scripts to be copied from the ForceBalance installation directory.
#         self.scripts = ['runcuda.sh']
#         ## Suffix for coordinate files.
#         self.crdsfx = '.pdb'
#         ## Command prefix.
#         self.prefix = "bash runcuda.sh"
#         if tgt_opts['remote_backup']:
#             self.prefix += " -b"
#         ## Initialize base class.
#         super(Hydration_OpenMM,self).__init__(options,tgt_opts,forcefield)
#         ## Send back the trajectory file.
#         if self.save_traj > 0:
#             self.extra_output = ['openmm-md.dcd']



from __future__ import absolute_import
import forcebalance
import shutil
from .test_target import TargetTests # general targets tests defined in test_target.py
"""
The testing functions for this class are located in test_target.py.
"""
class TestAbInitio_GMX(TargetTests):
    def setup_method(self, method):
        super(TestAbInitio_GMX, self).setup_method(method)
        self.options.update({
                'penalty_additive': 0.01,
                'jobtype': 'NEWTON',
                'forcefield': ['water.itp']})

        self.tgt_opt.update({'type':'ABINITIO_GMX',
                             'name':'cluster-02'})

        self.ff = forcebalance.forcefield.FF(self.options)

        self.ffname = self.options['forcefield'][0][:-3]
        self.filetype = self.options['forcefield'][0][-3:]
        self.mvals = [.5]*self.ff.np

        self.logger.debug("Setting up AbInitio_GMX target\n")
        self.target = forcebalance.gmxio.AbInitio_GMX(self.options, self.tgt_opt, self.ff)

    def teardown_method(self):
        shutil.rmtree('temp')
        super(TestAbInitio_GMX, self).teardown_method()




""" @package forcebalance.torsion_profile Torsion profile fitting module.

@author Lee-Ping Wang
@date 08/2019
"""
from __future__ import division
import os
import numpy as np
import itertools
import json
from copy import deepcopy
from collections import OrderedDict
from forcebalance.nifty import eqcgmx, printcool, printcool_dictionary, warn_press_key
from forcebalance.target import Target
from forcebalance.molecule import Molecule
from forcebalance.finite_difference import fdwrap, f12d3p, in_fd
from forcebalance.output import getLogger
from forcebalance.optimizer import Counter
logger = getLogger(__name__)

class TorsionProfileTarget(Target):
    """ Subclass of Target for fitting MM optimized geometries to QM optimized geometries. """
    def __init__(self, options, tgt_opts, forcefield):
        super(TorsionProfileTarget, self).__init__(options, tgt_opts, forcefield)

        if not os.path.exists(os.path.join(self.tgtdir, 'metadata.json')):
            raise RuntimeError('TorsionProfileTarget needs torsion drive metadata.json file')
        with open(os.path.join(self.tgtdir, 'metadata.json')) as f:
            self.metadata = json.load(f)
            self.ndim = len(self.metadata['dihedrals'])
            self.freeze_atoms = sorted(set(itertools.chain(*self.metadata['dihedrals'])))

        ## Read in the coordinate files and get topology information from PDB
        if hasattr(self, 'pdb') and self.pdb is not None:
            self.mol = Molecule(os.path.join(self.root,self.tgtdir,self.coords),
                                top=(os.path.join(self.root,self.tgtdir,self.pdb)))
        else:
            self.mol = Molecule(os.path.join(self.root,self.tgtdir,self.coords))
        ## Number of snapshots.
        self.ns = len(self.mol)
        ## Option for how much data to write to disk.
        self.set_option(tgt_opts,'writelevel','writelevel')
        ## Harmonic restraint for non-torsion atoms in kcal/mol.
        self.set_option(tgt_opts,'restrain_k','restrain_k')
        ## Attenuate the weights as a function of energy
        self.set_option(tgt_opts,'attenuate','attenuate')
        ## Energy denominator for objective function
        self.set_option(tgt_opts,'energy_denom','energy_denom')
        ## Set upper cutoff energy
        self.set_option(tgt_opts,'energy_upper','energy_upper')
        ## Read in the reference data.
        self.read_reference_data()
        ## Build keyword dictionaries to pass to engine.
        engine_args = OrderedDict(list(self.OptionDict.items()) + list(options.items()))
        engine_args.pop('name', None)
        ## Create engine object.
        engine_args['freeze_atoms'] = self.freeze_atoms
        self.engine = self.engine_(target=self, mol=self.mol, **engine_args)

    def read_reference_data(self):

        """ Read the reference ab initio data from a file such as qdata.txt.

        After reading in the information from qdata.txt, it is converted
        into the GROMACS energy units (kind of an arbitrary choice).
        """
        ## Reference (QM) energies
        self.eqm           = []
        ## The qdata.txt file that contains the QM energies and forces
        self.qfnm = os.path.join(self.tgtdir,"qdata.txt")
        # Parse the qdata.txt file
        for line in open(os.path.join(self.root,self.qfnm)):
            sline = line.split()
            if len(sline) == 0: continue
            elif sline[0] == 'ENERGY':
                self.eqm.append(float(sline[1]))

        if len(self.eqm) != self.ns:
            raise RuntimeError("Length of qdata.txt should match number of structures")

        # Turn everything into arrays, convert to kcal/mol
        self.eqm = np.array(self.eqm)
        self.eqm *= eqcgmx / 4.184
        # Use the minimum energy structure of the QM as reference
        self.eqm  -= np.min(self.eqm)
        self.smin  = np.argmin(self.eqm)
        logger.info("Referencing all energies to the snapshot %i (minimum energy structure in QM)\n" % self.smin)

        if self.attenuate:
            # Attenuate energies by an amount proportional to their
            # value above the minimum.
            eqm1 = self.eqm - np.min(self.eqm)
            denom = self.energy_denom
            upper = self.energy_upper
            self.wts = np.ones(self.ns)
            for i in range(self.ns):
                if eqm1[i] > upper:
                    self.wts[i] = 0.0
                elif eqm1[i] < denom:
                    self.wts[i] = 1.0 / denom
                else:
                    self.wts[i] = 1.0 / np.sqrt(denom**2 + (eqm1[i]-denom)**2)
        else:
            self.wts = np.ones(self.ns)

        # Normalize weights.
        self.wts /= sum(self.wts)

    def indicate(self):
        title_str = "Torsion Profile: %s, Objective = % .5e, Units = kcal/mol, Angstrom" % (self.name, self.objective)
        #LPW: This title is carefully placed to align correctly
        column_head_str1 =  "%-50s %-10s %-12s %-18s %-12s %-10s %-11s %-10s" % ("System", "Min(QM)", "Min(MM)", "Range(QM)", "Range(MM)", "Max-RMSD", "Ene-RMSE", "Obj-Fn")
        printcool_dictionary(self.PrintDict,title=title_str + '\n' + column_head_str1, keywidth=50, center=[True,False])

    def get(self, mvals, AGrad=False, AHess=False):
        Answer = {'X':0.0, 'G':np.zeros(self.FF.np), 'H':np.zeros((self.FF.np, self.FF.np))}
        self.PrintDict = OrderedDict()

        def compute(mvals_, indicate=False):
            self.FF.make(mvals_)
            M_opts = None
            compute.emm = []
            compute.rmsd = []
            for i in range(self.ns):
                energy, rmsd, M_opt = self.engine.optimize(shot=i, align=False)
                # Create a molecule object to hold the MM-optimized structures
                compute.emm.append(energy)
                compute.rmsd.append(rmsd)
                if M_opts is None:
                    M_opts = deepcopy(M_opt)
                else:
                    M_opts += M_opt
            compute.emm = np.array(compute.emm)
            compute.emm -= compute.emm[self.smin]
            compute.rmsd = np.array(compute.rmsd)
            if indicate:
                if self.writelevel > 0:
                    energy_comparison = np.array([
                        self.eqm,
                        compute.emm,
                        compute.emm - self.eqm,
                        np.sqrt(self.wts)/self.energy_denom
                    ]).T
                    np.savetxt("EnergyCompare.txt", energy_comparison, header="%11s  %12s  %12s  %12s" % ("QMEnergy", "MMEnergy", "Delta(MM-QM)", "Weight"), fmt="% 12.6e")
                    M_opts.write('mm_minimized.xyz')
                    if self.ndim == 1:
                        try:
                            import matplotlib.pyplot as plt
                            plt.switch_backend('agg')
                            fig, ax = plt.subplots()
                            dihedrals = np.array([i[0] for i in self.metadata['torsion_grid_ids']])
                            dsort = np.argsort(dihedrals)
                            ax.plot(dihedrals[dsort], self.eqm[dsort], label='QM')
                            if hasattr(self, 'emm_orig'):
                                ax.plot(dihedrals[dsort], compute.emm[dsort], label='MM Current')
                                ax.plot(dihedrals[dsort], self.emm_orig[dsort], label='MM Initial')
                            else:
                                ax.plot(dihedrals[dsort], compute.emm[dsort], label='MM Initial')
                                self.emm_orig = compute.emm.copy()
                            ax.legend()
                            ax.set_xlabel('Dihedral (degree)')
                            ax.set_ylabel('Energy (kcal/mol)')
                            fig.suptitle('Torsion profile: iteration %i\nSystem: %s' % (Counter(), self.name))
                            fig.savefig('plot_torsion.pdf')
                        except ImportError:
                            logger.warning("matplotlib package is needed to make torsion profile plots\n")
            return (np.sqrt(self.wts)/self.energy_denom) * (compute.emm - self.eqm)
        compute.emm = None
        compute.rmsd = None

        V = compute(mvals, indicate=True)

        Answer['X'] = np.dot(V,V)

        # Energy RMSE
        e_rmse = np.sqrt(np.dot(self.wts, (compute.emm - self.eqm)**2))

        self.PrintDict[self.name] = '%10s %10s    %6.3f - %-6.3f   % 6.3f - %-6.3f    %6.3f    %7.4f   % 7.4f' % (','.join(['%i' % i for i in self.metadata['torsion_grid_ids'][self.smin]]),
                                                                                                                  ','.join(['%i' % i for i in self.metadata['torsion_grid_ids'][np.argmin(compute.emm)]]),
                                                                                                                  min(self.eqm), max(self.eqm), min(compute.emm), max(compute.emm), max(compute.rmsd), e_rmse, Answer['X'])

        # compute gradients and hessian
        dV = np.zeros((self.FF.np,len(V)))
        if AGrad or AHess:
            for p in self.pgrad:
                dV[p,:], _ = f12d3p(fdwrap(compute, mvals, p), h = self.h, f0 = V)

        for p in self.pgrad:
            Answer['G'][p] = 2*np.dot(V, dV[p,:])
            for q in self.pgrad:
                Answer['H'][p,q] = 2*np.dot(dV[p,:], dV[q,:])
        if not in_fd():
            self.objective = Answer['X']
            self.FF.make(mvals)
        return Answer



from __future__ import absolute_import
from builtins import str
import pytest
import os
import forcebalance.molecule
from .__init__ import ForceBalanceTestCase
import numpy as np

class TestPDBMolecule(ForceBalanceTestCase):

    @classmethod
    def setup_class(cls):
        cls.source = 'dms_conf.pdb'
        super(TestPDBMolecule, cls).setup_class()

    def setup_method(self, method):
        super(TestPDBMolecule, self).setup_method(method)
        os.chdir('files')
        try: self.molecule = forcebalance.molecule.Molecule(self.source, build_topology=False)
        except IOError:
            self.skipTest("Input pdb file test/files/%s doesn't exist" % self.source)
        except:
            self.fail("\nUnable to open pdb file")

    def teardown_method(self):
        os.system('rm -rf {name}.xyz {name}.gro {name}.arc'.format(name=self.source[:-4]))
        super(TestPDBMolecule, self).teardown_method()

    def test_xyz_conversion(self):
        """Check molecule conversion from pdb to xyz format"""
        self.logger.debug("\nCreating xyz file from pdb... ")
        self.molecule.write(self.source[:-3] + 'xyz')
        self.logger.debug("done\nTrying to read generated xyz file... ")
        try:
            molecule1 = forcebalance.molecule.Molecule(self.source[:-3] + 'xyz', build_topology=False)
            self.logger.debug("ok\n")
        except:
            self.fail("\nConversion to xyz format creates unreadable file")

        self.logger.debug("Checking that conversion has not changed molecule spatial coordinates\n")
        assert (self.molecule.Data['xyzs'][0] == molecule1.Data['xyzs'][0]).all(), "Conversion from pdb to xyz yields "\
            "different xyz coordinates\npdb:\n%s\n\ngro:\n%s\n" %(str(self.molecule.Data['xyzs'][0]),
                                                                  str(molecule1.Data['xyzs'][0]))

    def test_measure_distances(self):
        """Check measure distance functions"""
        result = np.array(self.molecule.measure_distances(1701, 1706))
        EXPECTED_DISTANCE_RESULTS = np.array([1.80])
        np.testing.assert_allclose(EXPECTED_DISTANCE_RESULTS,result,atol=0.01,
                                   err_msg="\nMeasured distance has changed from previously calculated values.\n"
                                           "If this seems reasonable, update EXPECTED_DISTANCE_RESULTS in test_"
                                           "molecule.py with these values")

    def test_measure_angles(self):
        """Check measure angle functions"""
        result = np.array(self.molecule.measure_angles(1771, 1769, 1772))
        EXPECTED_ANGLE_RESULTS = np.array([114.11])
        np.testing.assert_allclose(EXPECTED_ANGLE_RESULTS,result,atol=0.01,
                                   err_msg="\nMeasured angle has changed from previously calculated values.\n"
                                           "If this seems reasonable, update EXPECTED_ANGLE_RESULTS in "
                                           "test_molecule.py with these values")

    def test_measure_dihedrals(self):
        """Check measure dihedral functions"""
        result = np.array(self.molecule.measure_dihedrals(1709, 1706, 1701, 1702))
        EXPECTED_DIHEDRAL_RESULTS = np.array([176.54])
        np.testing.assert_allclose(EXPECTED_DIHEDRAL_RESULTS,result,atol=0.01,
                                   err_msg="\nMeasured dihedral angle has changed from previously calculated values.\n"
                                           "If this seems reasonable, update EXPECTED_DIHEDRAL_RESULTS "
                                           "in test_molecule.py with these values")

    def test_gro_conversion(self):
        """Check molecule conversion from pdb to gro format"""
        self.logger.debug("\nCreating gro file from pdb... ")
        self.molecule.write(self.source[:-3] + 'gro')
        self.logger.debug("done\nTrying to read generated gro file... ")
        try:
            molecule1 = forcebalance.molecule.Molecule(self.source[:-3] + 'gro', build_topology=False)
            self.logger.debug("ok\n")
        except:
            self.fail("Conversion to gro format creates unreadable file")

        self.logger.debug("\nChecking that conversion has not changed number of residues\n")
        assert len(self.molecule.Data['resid']) == len(molecule1.Data['resid'])

        self.logger.debug("Checking that conversion has not changed molecule spatial coordinates\n")
        msg = "\nConversion from pdb to gro yields different xyz coordinates\npdb:\n%s\n\ngro:\n" \
              "%s\n" % (str(self.molecule.Data['xyzs'][0]), str(molecule1.Data['xyzs'][0]))
        np.testing.assert_allclose(self.molecule.Data['xyzs'][0],  molecule1.Data['xyzs'][0], rtol=0, atol=0.001,
                                   err_msg=msg)


    def test_arc_conversion(self):
        """Check molecule conversion from pdb to arc format"""
        self.logger.debug("\nCreating arc file from pdb... ")
        self.molecule.Data['tinkersuf']=['']*len(self.molecule.Data['resname'])  # suppress topology warning
        self.molecule.write(self.source[:-3] + 'arc')
        self.logger.debug("done\nTrying to read generated gro file... ")
        try:
            molecule1 = forcebalance.molecule.Molecule(self.source[:-3] + 'arc',build_topology=False)
            self.logger.debug("ok\n")
        except:
            self.fail("Conversion to arc (TINKER) format creates unreadable file")

        self.logger.debug("Checking that conversion has not changed molecule spatial coordinates\n")

        msg = "Conversion from pdb to arc (TINKER) yields different number of residues"
        assert len(self.molecule.Data['resid']) == len(molecule1.Data['resid']), msg

        msg = "\nConversion from pdb to arc yields different xyz coordinates" \
              "\npdb:\n%s\n\narc:\n%s\n" %(str(self.molecule.Data['xyzs'][0]),
                                           str(molecule1.Data['xyzs'][0]))
        np.testing.assert_allclose(self.molecule.Data['xyzs'][0], molecule1.Data['xyzs'][0],
                                   atol=0.000001, err_msg=msg)

    def test_pdb_topology_build(self):
        """Check reading pdb with build_topology=True"""
        self.logger.debug("\nTrying to read molecule with topology... ")
        try:
            molecule = forcebalance.molecule.Molecule(self.source, build_topology=True)
            self.logger.debug("\ndone\nChecking molecule has correct number of residues")
        except:
            self.fail("Failed to load pdb with build_topology=True")

        assert len(self.molecule.Data['resid']) == len(molecule.Data['resid']), "Topology build yields " \
                                                                                "different number of residues"

class TestLipidGRO(ForceBalanceTestCase):
    @classmethod
    def setup_class(cls):
        super(TestLipidGRO, cls).setup_class()
        cls.source = 'lipid.gro'

    def setup_method(self, method):
        super(TestLipidGRO, self).setup_method(method)
        os.chdir('files')
        try: self.molecule = forcebalance.molecule.Molecule(self.source, toppbc=True)
        except IOError:
            pytest.skip("Input pdb file test/files/%s doesn't exist" % self.source)
        except:
            pytest.fail("Unable to open gro file")
            
    def test_measure_dihedrals(self):
        """Check measure dihedral functions"""
        result = np.array(self.molecule.measure_dihedrals(1131, 1112, 1113, 1114))
        EXPECTED_DIHEDRAL_RESULTS = np.array([-157.223])
        msg = "\nMeasured dihedral angle has changed from previously calculated values.\n" \
              "If this seems reasonable, update EXPECTED_DIHEDRAL_RESULTS in test_molecule.py with these values"
        np.testing.assert_allclose(EXPECTED_DIHEDRAL_RESULTS, result, atol=0.001, err_msg=msg)


    def test_lipid_molecules(self):
        """Check for the correct number of molecules in a rectangular cell with broken molecules"""
        self.logger.debug("\nTrying to read lipid conformation... ")
        assert len(self.molecule.molecules) == 3783, "Incorrect number of molecules for lipid structure"

class TestWaterPDB(ForceBalanceTestCase):
    @classmethod
    def setup_class(cls):
        super(TestWaterPDB, cls).setup_class()
        cls.source = 'waterbox500.pdb'

    def setup_method(self, method):
        super(TestWaterPDB, self).setup_method(method)
        os.chdir('files')
        try: self.molecule = forcebalance.molecule.Molecule(self.source, toppbc=True)
        except IOError:
            pytest.skip("Input pdb file test/files/%s doesn't exist" % self.source)
        except:
            pytest.fail("Unable to open pdb file")

    def test_water_molecules(self):
        """Check for the correct number of molecules in a cubic water box"""
        self.logger.debug("\nTrying to read water conformation... ")
        assert len(self.molecule.molecules) == 500, "Incorrect number of molecules for water structure"

class TestAlaGRO(ForceBalanceTestCase):
    @classmethod
    def setup_class(cls):
        super(TestAlaGRO, cls).setup_class()
        cls.source = 'ala.gro'

    def setup_method(self, method):
        super(TestAlaGRO, self).setup_method(method)
        os.chdir('files')
        try: self.molecule = forcebalance.molecule.Molecule(self.source)
        except IOError:
            pytest.skip("Input gro file test/files/%s doesn't exist" % self.source)
        except:
            pytest.fail("Unable to open gro file")

    def test_ala_molecules(self):
        """Check for the correct number of bonds in a simple molecule"""
        self.logger.debug("\nTrying to read alanine dipeptide conformation... ")
        assert len(self.molecule.bonds) == 21, "Incorrect number of bonds for alanine dipeptide structure"


class TestGalbPNPMol2(ForceBalanceTestCase):
    @classmethod
    def setup_class(cls):
        super(TestGalbPNPMol2, cls).setup_class()
        cls.source = 'pNP-0LB-tleap.mol2'

    def setup_method(self, method):
        super(TestGalbPNPMol2, self).setup_method(method)
        os.chdir('files')
        try: self.molecule = forcebalance.molecule.Molecule(self.source)
        except IOError:
            pytest.skip("Input gro file test/files/%s doesn't exist" % self.source)
        except:
            pytest.fail("Unable to open mol2 file")

    def test_read_galb(self):
        """Check for the correct number of bonds in a simple molecule"""
        self.logger.debug("\nTrying to read alanine dipeptide conformation... ")
        assert self.molecule.resname == 14 * ['PNP'] + 22 * ['0LB'], "Incorrect residue names"
        assert self.molecule.elem == ['O', 'N', 'O', 'C', 'C', 'C', 'H', 'H', 'C', 'H', 'C', 'H', 'C', 'O', 'C', 'H',
                                      'O', 'C', 'H', 'C', 'H', 'H', 'O', 'H', 'C', 'H', 'O', 'H', 'C', 'H', 'O', 'H',
                                      'C', 'H', 'O', 'H'], "Incorrect atomic symbols"
        assert len(self.molecule.bonds) == 37, "Incorrect number of bonds for pNP-0LB structure"



from __future__ import division
from __future__ import absolute_import
from builtins import str
from .__init__ import ForceBalanceTestCase
import pytest
import forcebalance
from forcebalance.nifty import *
from forcebalance.nifty import _exec

try:
    import work_queue
except ImportError:
    work_queue = None

class TestNifty(ForceBalanceTestCase):
    def setup_method(self, method):
        super(TestNifty, self).setup_method(method)
        # skip work_queue tests if work_queue could not be imported
        if re.match(".*work_queue.*", method.__name__) and not work_queue:
                pytest.skip("work_queue module not installed")

        # Maintaining a cleanup_func list allows us to add tasks to the teardown process during the test iteself
        self.cleanup_funcs = list()

    def teardown_method(self):
        # Maintaining a cleanup_func list allows us to add tasks to the teardown process during the test iteself
        for cleanup_func in self.cleanup_funcs:
            cleanup_func()

    def test_nifty_functions(self):
        """Check utility functions in forcebalance.nifty"""

        ##variable manipulation functions
        self.logger.debug("Checking nifty.isint()\n")
        assert isint("1")
        assert not(isint("1."))
        assert isint("-4")
        assert not(isint("-3.14"))

        self.logger.debug("Checking nifty.isfloat()\n")
        assert isfloat("1.5")
        assert isfloat("1")
        assert not(isfloat("a"))

        self.logger.debug("Checking nifty.isdecimal()\n")
        assert isdecimal("1.0")
        assert not(isdecimal("1"))

        for result in get_least_squares(([0]), [0]):
            assert not(result.any())

        self.logger.debug("Verifying nifty.get_least_squares() results for some trivial cases\n")
        ##least squares function tests
        #   trivial fully determined
        X=((1,3,-2),(3,5,6),(2,4,3))
        Y=(5,7,8)
        result = get_least_squares(X,Y)[0]
        np.testing.assert_almost_equal(result[0], -15)
        np.testing.assert_almost_equal(result[1], 8)
        np.testing.assert_almost_equal(result[2], 2)

        #   inconsistent system
        X=((1,),(1,))
        Y=(0,1)
        result = get_least_squares(X,Y)[0]
        np.testing.assert_almost_equal(result[0], .5)

        #   overdetermined system
        X=((2,0),(-1,1),(0,2))
        Y=(1,0,-1)
        result = get_least_squares(X,Y)[0]
        np.testing.assert_almost_equal(result[0], 1./3.)
        np.testing.assert_almost_equal(result[1], -1./3.)

        self.logger.debug("Verify nifty matrix manipulations perform as expected\n")
        ##matrix manipulations
        X=flat(X)
        assert X.shape == (6,)
        X=row(X)
        assert X.shape == (1,6)
        X=col(X)
        assert X.shape == (6,1)

        self.logger.debug("Running some test processes using nifty._exec()\n")
        ##_exec
        assert type(_exec("")) is list
        assert _exec("echo test")[0] == "test"
        _exec("touch .test")
        assert os.path.isfile(".test")
        _exec("rm .test")
        assert not(os.path.isfile(".test"))
        with pytest.raises(Exception) as excinfo:
            _exec("exit 255")

    def test_work_queue_functions(self):
        """Check work_queue functions behave as expected"""
        
        # Work Queue will no longer be initialized to None
        self.logger.debug("\nChecking Work Queue is initialized to None...\n")
        assert forcebalance.nifty.WORK_QUEUE is None, "Unexpected initialization of forcebalance.nifty.WORK_QUEUE " \
                                                      "to %s" % str(forcebalance.nifty.WORK_QUEUE)
        self.logger.info("\n")

        createWorkQueue(9191, debug=False)
        self.logger.debug("Created work queue, verifying...\n")
        assert type(forcebalance.nifty.WORK_QUEUE) is work_queue.WorkQueue, "Expected forcebalance.nifty.WORK_QUEUE to " \
                                                                            "be a WorkQueue object, but got a %s " \
                                                                            "instead" % str(type(forcebalance.nifty.WORK_QUEUE))
        self.logger.debug("Checking that getWorkQueue() returns valid WorkQueue object...\n")
        wq = getWorkQueue()
        assert type(wq) is work_queue.WorkQueue, "Expected getWorkQueue() to return a " \
                                                 "WorkQueue object, but got %s instead" % str(type(wq))
        worker_program = which('work_queue_worker')
        if worker_program != '':
            self.logger.debug("Submitting test job 'echo work queue test > test.job'\n")
            queue_up(wq, "echo work queue test > test.job", [], ["test.job"], tgt=None, verbose=False)
            self.logger.debug("Verifying that work queue has a task waiting\n")
            assert wq.stats.tasks_waiting == 1, "Expected queue to have a task waiting"
            
            self.logger.debug("Creating work_queue_worker process... ")
            worker = subprocess.Popen([os.path.join(worker_program, "work_queue_worker"),
                                       "localhost",
                                       str(wq.port)],
                                      stdout=subprocess.PIPE)
            #self.addCleanup(worker.terminate)
            self.cleanup_funcs.append(worker.terminate)
            self.logger.debug("Done\nTrying to get task from work queue\n")
            
            self.logger.debug("Calling wq_wait1 to fetch task\n")
            wq_wait1(wq, wait_time=5)
            self.logger.debug("wq_wait1(wq, wait_time=5) finished\n")
            self.logger.debug("Checking that wq.stats.total_tasks_complete == 1\n")
            # self.assertEqual(wq.stats.total_tasks_complete, 1, msg = "\nExpected queue to have a task completed")
            assert wq.stats.total_tasks_complete == 1, "Expected queue to have a task completed"
        else:
            self.logger.debug("work_queue_worker is not in the PATH.\n")
        
        # Destroy the Work Queue object so it doesn't interfere with the rest of the tests.
        destroyWorkQueue()



## HACK: Improve the performance of the openff forcefield.create_openmm_system()

# time based on total 540s evaluation
# cache for find_smarts_matches (save 300+ s)
from openforcefield.utils.toolkits import OpenEyeToolkitWrapper, RDKitToolkitWrapper
original_find_smarts_matches = OpenEyeToolkitWrapper.find_smarts_matches
TOOLKIT_CACHE_find_smarts_matches = {}
def cached_find_smarts_matches(self, molecule, smarts, aromaticity_model='OEAroModel_MDL'):
    cache_key = hash((molecule, smarts, aromaticity_model))
    if cache_key not in TOOLKIT_CACHE_find_smarts_matches:
        TOOLKIT_CACHE_find_smarts_matches[cache_key] = original_find_smarts_matches(self, molecule, smarts, aromaticity_model=aromaticity_model)
    return TOOLKIT_CACHE_find_smarts_matches[cache_key]
# replace the original function with new one
OpenEyeToolkitWrapper.find_smarts_matches = cached_find_smarts_matches


# cache for the validate function (save 94s)
from openforcefield.typing.chemistry.environment import ChemicalEnvironment
original_validate = ChemicalEnvironment.validate
TOOLKIT_CACHE_ChemicalEnvironment_validate = {}
def cached_validate(smirks, ensure_valence_type=None, toolkit='openeye'):
    cache_key = hash((smirks, ensure_valence_type, toolkit))
    if cache_key not in TOOLKIT_CACHE_ChemicalEnvironment_validate:
        TOOLKIT_CACHE_ChemicalEnvironment_validate[cache_key] = original_validate(smirks, ensure_valence_type=None, toolkit=toolkit)
    return TOOLKIT_CACHE_ChemicalEnvironment_validate[cache_key]
ChemicalEnvironment.validate = cached_validate


# cache for compute_partial_charges_am1bcc (save 69s)
original_compute_partial_charges_am1bcc = OpenEyeToolkitWrapper.compute_partial_charges_am1bcc
TOOLKIT_CACHE_compute_partial_charges_am1bcc = {}
def cached_compute_partial_charges_am1bcc(self, molecule):
    cache_key = hash(molecule)
    if cache_key not in TOOLKIT_CACHE_compute_partial_charges_am1bcc:
        TOOLKIT_CACHE_compute_partial_charges_am1bcc[cache_key] = original_compute_partial_charges_am1bcc(self, molecule)
    return TOOLKIT_CACHE_compute_partial_charges_am1bcc[cache_key]
OpenEyeToolkitWrapper.compute_partial_charges_am1bcc = cached_compute_partial_charges_am1bcc

# cache the generate_conformers function (save 15s)
TOOLKIT_CACHE_molecule_conformers = {}
original_generate_conformers = OpenEyeToolkitWrapper.generate_conformers
def cached_generate_conformers(self, molecule, n_conformers=1, clear_existing=True):
    cache_key = hash((molecule, n_conformers, clear_existing))
    if cache_key not in TOOLKIT_CACHE_molecule_conformers:
        original_generate_conformers(self, molecule, n_conformers=n_conformers, clear_existing=clear_existing)
        TOOLKIT_CACHE_molecule_conformers[cache_key] = molecule._conformers
    molecule._conformers = TOOLKIT_CACHE_molecule_conformers[cache_key]
OpenEyeToolkitWrapper.generate_conformers = cached_generate_conformers

# final timing: 56s

# cache the ForceField creation (no longer needed since using OpenFF API for parameter modifications)

import hashlib
from openforcefield.typing.engines.smirnoff import ForceField
SMIRNOFF_FORCE_FIELD_CACHE = {}
def getForceField(*ffpaths):
    hasher = hashlib.md5()
    for path in ffpaths:
        with open(path, 'rb') as f:
            hasher.update(f.read())
    cache_key = hasher.hexdigest()
    if cache_key not in SMIRNOFF_FORCE_FIELD_CACHE:
        SMIRNOFF_FORCE_FIELD_CACHE[cache_key] = ForceField(*ffpaths, allow_cosmetic_attributes=True)
    return SMIRNOFF_FORCE_FIELD_CACHE[cache_key]



from __future__ import division
from __future__ import absolute_import
from builtins import str
from builtins import range
import os
import forcebalance
import numpy
from .__init__ import ForceBalanceTestCase

class TargetTests(ForceBalanceTestCase):

    def setup_method(self, method):
        super(TargetTests, self).setup_method(method)
        self.logger = forcebalance.output.getLogger('forcebalance.test.' + __name__[5:])
        self.logger.debug("\nBuilding options for target...\n")
        self.options = forcebalance.parser.gen_opts_defaults.copy()
        self.tgt_opt = forcebalance.parser.tgt_opts_defaults.copy()
        self.ff = None  # Forcefield this target is fitting
        self.options.update({'root': os.path.join(os.getcwd(), 'files')})

        os.chdir(self.options['root'])

    def test_get_function(self):
        """Check target get() function output"""
        #os.chdir(self.target.tempdir)
        os.chdir('temp/%s' % self.tgt_opt['name'])


        self.logger.debug("Evaluating objective function for target...\n")
        objective = self.target.get(self.mvals)
        self.target.indicate()
        print("objective =\n%s" % str(objective))

        # check objective dictionary keys
        print("\n>ASSERT objective dictionary has X, G, H keys\n")
        assert isinstance(objective, dict)
        assert 'X' in objective
        assert 'G' in objective
        assert 'H' in objective

        # check objective value types
        print(">ASSERT objective['X'] is a float\n")
        assert isinstance(objective['X'], numpy.float64)
        print(">ASSERT objective['G'] is a numpy array\n")
        assert isinstance(objective['G'], numpy.ndarray)
        print(">ASSERT objective['H'] is a numpy array\n")
        assert isinstance(objective['H'], numpy.ndarray)

        # check array dimensions
        print(">ASSERT size of objective['G'] is a equal to number of forcefield parameters (p)\n")
        assert objective['G'].size == self.ff.np
        print(">ASSERT size of objective['H'] is a equal to number of forcefield parameters squared (p^2)\n")
        assert objective['H'].size == self.ff.np**2
        print(">ASSERT objective['G'] is one dimensional\n")
        assert objective['G'].ndim == 1
        print(">ASSERT objective['H'] is two dimensional\n")
        assert objective['H'].ndim == 2
        print(">ASSERT objective['G'] is p x 1 array\n")
        assert objective['G'].shape == (self.ff.np,)
        print(">ASSERT objective['G'] is p x p array\n")
        assert objective['H'].shape == (self.ff.np, self.ff.np)

        os.chdir('../..')

    def test_get_agrad(self):
        """Check target objective function gradient using finite difference"""
        self.mvals = [.5]*self.ff.np

        os.chdir(os.path.join('temp', self.tgt_opt['name']))

        print("Running target.get(mvals, AGrad=True)\n")
        objective = self.target.get(self.mvals, AGrad=True)
        X = objective['X']
        G = objective['G']
        print(">ASSERT objective['G'] is not a zero vector\n")
        assert G.any() # with AGrad=True, G should not be [0]
        g = numpy.zeros(self.ff.np)

        print(">ASSERT objective['G'] approximately matches finite difference calculations\n")
        for p in range(self.ff.np):
            mvals_lo = self.mvals[:]
            mvals_hi = self.mvals[:]
            mvals_lo[p]-=(self.mvals[p]/200.)
            mvals_hi[p]+=(self.mvals[p]/200.)

            Xlo = self.target.get(mvals_lo)['X']
            Xhi = self.target.get(mvals_hi)['X']
            g[p] = (Xhi-Xlo)/(self.mvals[p]/100.)
            assert abs(g[p]-G[p]) < X*.01 +1e-7

        os.chdir('../..')



import pytest
from collections import defaultdict

def pytest_runtest_makereport(item, call):
    if item.cls is None: return
    if not hasattr(item.cls, 'failcount'):
        item.cls.failcount = 0
    if call.excinfo is not None:
        item.cls.failcount += 1
    # print("LPW: Running pytest_runtest_makereport", item.name, item.parent.name)
    # import IPython
    # IPython.embed()
    # if "incremental" in item.keywords:
    #     if call.excinfo is not None:
    #         parent = item.parent
    #         parent._previousfailed = item

# def pytest_runtest_setup(item):
#     previousfailed = getattr(item.parent, "_previousfailed", None)
#     if previousfailed is not None:
#         pytest.xfail("previous test failed (%s)" % previousfailed.name)



import os
import shutil
import forcebalance.parser
from .__init__ import ForceBalanceTestCase

class TestParser(ForceBalanceTestCase):

    def test_parse_inputs_returns_tuple(self):
        """Check parse_inputs() returns type"""
        output = forcebalance.parser.parse_inputs('files/very_simple.in')
        assert isinstance(output, tuple), "\nExpected parse_inputs() to return a tuple, but got a %s instead" % type(output).__name__
        assert isinstance(output[0], dict), "\nExpected parse_inputs()[0] to be an options dictionary, got a %s instead" % type(output).__name__
        assert isinstance(output[1], list), "\nExpected parse_inputs()[1] to be a target list, got a %s instead" % type(output[1]).__name__
        assert isinstance(output[1][0], dict), "\nExpected parse_inputs()[1][0] to be a target dictionary, got a %s instead" % type(output[1]).__name__

    def test_parse_inputs_generates_default_options(self):
        """Check parse_inputs() without arguments generates dictionary of default options"""
        output = forcebalance.parser.parse_inputs()
        defaults = forcebalance.parser.gen_opts_defaults
        defaults.update({'root':os.getcwd()})
        defaults.update({'input_file':None})
        target_defaults = forcebalance.parser.tgt_opts_defaults
        assert output[0] == defaults, "\nparse_inputs() target options do not match those in forcebalance.parser.gen_opts_defaults"
        assert output[1][0] == target_defaults, "\nparse_inputs() target options do not match those in forcebalance.parser.tgt_opts_defaults"

    def test_parse_inputs_yields_consistent_results(self):
        """Check parse_inputs() gives consistent results"""
        output1 = forcebalance.parser.parse_inputs('files/very_simple.in')
        output2 = forcebalance.parser.parse_inputs('files/very_simple.in')
        assert output1 == output2
        os.chdir('files')
        output3 = forcebalance.parser.parse_inputs('very_simple.in')
        output4 = forcebalance.parser.parse_inputs('very_simple.in')
        assert output3 == output4
        # directory change should lead to different result in output['root']
        assert output1 != output3
        # different parameters from the same file should yield different results
        shutil.copyfile('0.energy_force.in', 'test.in')
        output5 = forcebalance.parser.parse_inputs('test.in')
        shutil.copyfile('1.netforce_torque.in','test.in')
        output6 = forcebalance.parser.parse_inputs('test.in')
        assert output5 != output6
        os.remove('test.in')



""" @package forcebalance.evaluator
A target which employs the OpenFF Evaluator framework to compute condensed
phase physical properties. Currently only force fields in the  OpenFF SMIRNOFF
format are supported.

author Yudong Qiu, Simon Boothroyd, Lee-Ping Wang
@date 06/2019
"""
from __future__ import division, print_function

import json
import os
import re
import tempfile

import numpy as np
from forcebalance.nifty import warn_once, printcool, printcool_dictionary
from forcebalance.output import getLogger
from forcebalance.target import Target

try:
    from evaluator import unit
    from evaluator.client import EvaluatorClient, ConnectionOptions, RequestOptions
    from evaluator.datasets import PhysicalPropertyDataSet
    from evaluator.utils.exceptions import EvaluatorException
    from evaluator.utils.openmm import openmm_quantity_to_pint
    from evaluator.utils.serialization import TypedJSONDecoder, TypedJSONEncoder
    from evaluator.forcefield import ParameterGradientKey
except ImportError:
    warn_once("Failed to import the evaluator package.")

try:
    from openforcefield.typing.engines import smirnoff
except ImportError:
    warn_once("Failed to import the openforcefield package.")

logger = getLogger(__name__)


class Evaluator_SMIRNOFF(Target):
    """A custom optimisation target which employs the `openff-evaluator`
    package to rapidly estimate a collection of condensed phase physical
    properties at each optimisation epoch."""

    class OptionsFile:
        """Represents the set of options that a `Evaluator_SMIRNOFF`
        target will be run with.

        Attributes
        ----------
        connection_options: evaluator.client.ConnectionOptions
            The options for how the `evaluator` client should
            connect to a running server instance.
        estimation_options: evaluator.client.RequestOptions
            The options for how properties should be estimated by the
            `evaluator` (e.g. the uncertainties to which properties
            should be estimated).
        data_set_path: str
            The path to a JSON serialized PhysicalPropertyDataSet which
            contains those physical properties which will be optimised
            against.
        weights: dict of float
            The weighting of each property which will be optimised against.
        denominators: dict of str and unit.Quantity
            The denominators will be used to remove units from the properties
            and scale their values.
        """

        def __init__(self):

            self.connection_options = ConnectionOptions()
            self.estimation_options = RequestOptions()

            self.data_set_path = ""
            self.weights = {}
            self.denominators = {}

        def to_json(self):
            """Converts this class into a JSON string.

            Returns
            -------
            str
                The JSON representation of this class.
            """

            value = {
                "connection_options": self.connection_options.__getstate__(),
                "estimation_options": self.estimation_options.__getstate__(),
                "data_set_path": self.data_set_path,
                "weights": {
                    property_name: self.weights[property_name]
                    for property_name in self.weights
                },
                "denominators": {
                    property_name: self.denominators[property_name]
                    for property_name in self.denominators
                },
            }

            return json.dumps(
                value,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
                cls=TypedJSONEncoder,
            )

        @classmethod
        def from_json(cls, json_source):
            """Creates this class from a JSON string.

            Parameters
            -------
            json_source: str or file-like object
                The JSON representation of this class.
            """

            if isinstance(json_source, str):
                with open(json_source, "r") as file:
                    dictionary = json.load(file, cls=TypedJSONDecoder)
            else:
                dictionary = json.load(json_source, cls=TypedJSONDecoder)

            assert (
                "connection_options" in dictionary
                and "estimation_options" in dictionary
                and "data_set_path" in dictionary
                and "weights" in dictionary
                and "denominators" in dictionary
            )

            value = cls()

            value.connection_options = ConnectionOptions()
            value.connection_options.__setstate__(dictionary["connection_options"])

            value.estimation_options = RequestOptions()
            value.estimation_options.__setstate__(dictionary["estimation_options"])

            value.data_set_path = dictionary["data_set_path"]

            value.weights = {
                property_name: dictionary["weights"][property_name]
                for property_name in dictionary["weights"]
            }
            value.denominators = {
                property_name: dictionary["denominators"][property_name]
                for property_name in dictionary["denominators"]
            }

            return value

    def __init__(self, options, tgt_opts, forcefield):

        super(Evaluator_SMIRNOFF, self).__init__(options, tgt_opts, forcefield)

        self._options = None  # The options for this target loaded from JSON.
        self._default_units = (
            {}
        )  # The default units to convert each type of property to.

        self._client = None  # The client object which will communicate with an already spun up server.

        self._reference_data_set = None  # The data set of properties to estimate.
        self._normalised_weights = (
            None  # The normalised weights of the different properties.
        )

        # Store a `Future` like object which can be queried for the results of
        # a property estimation.
        self._pending_estimate_request = None

        # Store a mapping between gradient keys and the force balance string representation.
        self._gradient_key_mappings = {}
        self._parameter_units = {}

        # Store a copy of the objective function details from the previous optimisation cycle.
        self._last_obj_details = {}

        # Get the filename for the evaluator input file.
        self.set_option(tgt_opts, "evaluator_input", forceprint=True)

        # Initialize the target.
        self._initialize()

    def _initialize(self):
        """Initializes the evaluator target from an input json file.

        1. Reads the user specified input file.
        2. Creates a `evaluator` client object.
        3. Loads in a reference experimental data set.
        4. Assigns and normalises weights for each property.
        """

        # Load in the options from a user provided JSON file.
        print(os.path.join(self.tgtdir, self.evaluator_input))
        options_file_path = os.path.join(self.tgtdir, self.evaluator_input)
        self._options = self.OptionsFile.from_json(options_file_path)

        for property_type, denominator in self._options.denominators.items():
            self._default_units[property_type] = denominator.units

        # Attempt to create an evaluator client object using the specified
        # connection options.
        self._client = EvaluatorClient(self._options.connection_options)

        # Load in the experimental data set.
        data_set_path = os.path.join(self.tgtdir, self._options.data_set_path)
        self._reference_data_set = PhysicalPropertyDataSet.from_json(data_set_path)

        if len(self._reference_data_set) == 0:

            raise ValueError(
                "The physical property data set to optimise against is empty."
            )

        # Print the reference data, and count the number of instances of
        # each property type.
        printcool("Loaded experimental data.")

        property_types = self._reference_data_set.property_types

        number_of_properties = {
            x: sum(1 for y in self._reference_data_set.properties_by_type(x))
            for x in property_types
        }

        for substance in self._reference_data_set.substances:

            dict_for_print = {}

            for physical_property in self._reference_data_set.properties_by_substance(
                substance
            ):

                property_type = physical_property.__class__.__name__

                value = physical_property.value.to(self._default_units[property_type])
                uncertainty = physical_property.uncertainty.to(
                    self._default_units[property_type]
                )

                tuple_key = (
                    property_type,
                    physical_property.thermodynamic_state.temperature,
                    physical_property.thermodynamic_state.pressure,
                )

                dict_for_print["%s %s-%s" % tuple_key] = "%s+/-%s" % (
                    value,
                    uncertainty,
                )

            printcool_dictionary(
                dict_for_print, title="Reference %s data" % substance.identifier,
            )

        # Assign and normalize weights for each phase point (average for now)
        self._normalised_weights = {}

        for property_type in self._reference_data_set.property_types:

            self._normalised_weights[property_type] = (
                self._options.weights[property_type]
                / number_of_properties[property_type]
            )

    def _parameter_value_from_gradient_key(self, gradient_key):
        """Extracts the value of the parameter in the current
        open force field object pointed to by a given
        `ParameterGradientKey` object.

        Parameters
        ----------
        gradient_key: evaluator.forcefield.ParameterGradientKey
            The gradient key which points to the parameter of interest.

        Returns
        -------
        unit.Quantity
            The value of the parameter.
        bool
            Returns True if the parameter is a cosmetic one.
        """

        parameter_handler = self.FF.openff_forcefield.get_parameter_handler(
            gradient_key.tag
        )
        parameter = parameter_handler.parameters[gradient_key.smirks]

        attribute_split = re.split(r"(\d+)", gradient_key.attribute)
        attribute_split = list(filter(None, attribute_split))

        parameter_attribute = None
        parameter_value = None

        if hasattr(parameter, gradient_key.attribute):

            parameter_attribute = gradient_key.attribute
            parameter_value = getattr(parameter, parameter_attribute)

        elif len(attribute_split) == 2:

            parameter_attribute = attribute_split[0]

            if hasattr(parameter, parameter_attribute):
                parameter_index = int(attribute_split[1]) - 1

                parameter_value_list = getattr(parameter, parameter_attribute)
                parameter_value = parameter_value_list[parameter_index]

        is_cosmetic = False

        if (
            parameter_attribute is None
            or parameter_attribute in parameter._cosmetic_attribs
        ):
            is_cosmetic = True

        return openmm_quantity_to_pint(parameter_value), is_cosmetic

    def _extract_physical_parameter_values(self):
        """Extracts an array of the values of the physical parameters
        (which are not cosmetic) from the current `FF.openff_forcefield`
        object.

        Returns
        -------
        np.ndarray
            The array of values of shape (len(self._gradient_key_mappings),)
        """

        parameter_values = np.zeros(len(self._gradient_key_mappings))

        for gradient_key, parameter_index in self._gradient_key_mappings.items():

            parameter_value, _ = self._parameter_value_from_gradient_key(gradient_key)
            expected_unit = self._parameter_units[gradient_key]

            parameter_values[parameter_index] = parameter_value.to(
                expected_unit
            ).magnitude

        return parameter_values

    def _build_pvals_jacobian(self, mvals, perturbation_amount=1.0e-4):
        """Build the matrix which maps the gradients of properties with
        respect to physical parameters to gradients with respect to
        force balance mathematical parameters.

        Parameters
        ----------
        mvals: np.ndarray
            The current force balance mathematical parameters.
        perturbation_amount: float
            The amount to perturb the mathematical parameters by
            when calculating the finite difference gradients.

        Returns
        -------
        np.ndarray
            A matrix of d(Physical Parameter)/d(Mathematical Parameter).
        """

        jacobian_list = []

        for index in range(len(mvals)):

            reverse_mvals = mvals.copy()
            reverse_mvals[index] -= perturbation_amount

            self.FF.make(reverse_mvals)
            reverse_physical_values = self._extract_physical_parameter_values()

            forward_mvals = mvals.copy()
            forward_mvals[index] += perturbation_amount

            self.FF.make(forward_mvals)
            forward_physical_values = self._extract_physical_parameter_values()

            gradients = (forward_physical_values - reverse_physical_values) / (
                2.0 * perturbation_amount
            )
            jacobian_list.append(gradients)

        # Make sure to restore the FF object back to its original state.
        self.FF.make(mvals)

        jacobian = np.array(jacobian_list)
        return jacobian

    def submit_jobs(self, mvals, AGrad=True, AHess=True):
        """
        Submit jobs for evaluating the objective function

        Parameters
        ----------
        mvals: np.ndarray
            mvals array containing the math values of the parameters
        AGrad: bool
            Flag for computing gradients of not
        AHess: bool
            Flag for computing hessian or not

        Notes
        -----
        1. This function is called before wq_complete() and get().
        2. This function should not block.
        """

        # Make the force field based on the current values of the parameters.
        self.FF.make(mvals)

        force_field = smirnoff.ForceField(
            self.FF.offxml, allow_cosmetic_attributes=True
        )

        # strip out cosmetic attributes
        with tempfile.NamedTemporaryFile(mode="w", suffix=".offxml") as file:
            force_field.to_file(file.name, discard_cosmetic_attributes=True)
            force_field = smirnoff.ForceField(file.name)

        # Determine which gradients (if any) we should be estimating.
        parameter_gradient_keys = []

        self._gradient_key_mappings = {}
        self._parameter_units = {}

        if AGrad is True:

            index_counter = 0

            for field_list in self.FF.pfields:

                string_key = field_list[0]
                key_split = string_key.split("/")

                parameter_tag = key_split[0].strip()
                parameter_smirks = key_split[3].strip()
                parameter_attribute = key_split[2].strip()

                # Use the full attribute name (e.g. k1) for the gradient key.
                parameter_gradient_key = ParameterGradientKey(
                    tag=parameter_tag,
                    smirks=parameter_smirks,
                    attribute=parameter_attribute,
                )

                # Find the unit of the gradient parameter.
                parameter_value, is_cosmetic = self._parameter_value_from_gradient_key(
                    parameter_gradient_key
                )

                if parameter_value is None or is_cosmetic:
                    # We don't wan't gradients w.r.t. cosmetic parameters.
                    continue

                parameter_unit = parameter_value.units
                parameter_gradient_keys.append(parameter_gradient_key)

                self._gradient_key_mappings[parameter_gradient_key] = index_counter
                self._parameter_units[parameter_gradient_key] = parameter_unit

                index_counter += 1

        # Submit the estimation request.
        self._pending_estimate_request, _ = self._client.request_estimate(
            property_set=self._reference_data_set,
            force_field_source=force_field,
            options=self._options.estimation_options,
            parameter_gradient_keys=parameter_gradient_keys,
        )

        logger.info(
            "Requesting the estimation of {} properties, and their "
            "gradients with respect to {} parameters.\n".format(
                len(self._reference_data_set), len(parameter_gradient_keys)
            )
        )

        if self._pending_estimate_request.results(True)[0] is None:

            raise RuntimeError(
                "No `EvaluatorServer` could be found to submit the calculations to. "
                "Please double check that a server is running, and that the connection "
                "settings specified in the input script are correct."
            )

    @staticmethod
    def _check_estimation_request(estimation_request):
        """Checks whether an estimation request has finished with any exceptions.

        Parameters
        ----------
        estimation_request: evaluator.client.Request
            The request to check.
        """
        results, _ = estimation_request.results()

        if results is None:
            raise ValueError("Trying to extract the results of an unfinished request.")

        # Check for any exceptions that were raised while estimating
        # the properties.
        if isinstance(results, EvaluatorException):

            raise ValueError(
                "An uncaught exception occured within the evaluator "
                "framework: %s" % str(results)
            )

        if len(results.unsuccessful_properties) > 0:

            exceptions = "\n".join(str(result) for result in results.exceptions)

            raise ValueError(
                "Some properties could not be estimated:\n\n%s." % exceptions
            )

    def _extract_property_data(self, estimation_request, mvals, AGrad):
        """Extract the property estimates #and their gradients#
        from a relevant evaluator request object.

        Parameters
        ----------
        estimation_request: evaluator.client.Request
            The request to extract the data from.

        Returns
        -------
        estimated_data: evaluator.datasets.PhysicalPropertyDataSet
            The data set of estimated properties.
        estimated_gradients: dict of str and np.array
            The estimated gradients in a dictionary with keys of the estimated
            properties unique ids, and values of the properties gradients of shape
            (n_params,).
        """
        # Make sure the request actually finished and was error free.
        Evaluator_SMIRNOFF._check_estimation_request(estimation_request)

        # Extract the results from the request.
        results, _ = estimation_request.results()

        # Save a copy of the results to the temporary directory
        results_path = os.path.join(self.root, self.rundir, "results.json")
        results.json(results_path)

        # Print out some statistics about the calculation
        calculation_layer_counts = {}

        for physical_property in results.estimated_properties:

            calculation_layer = physical_property.source.fidelity

            if calculation_layer not in calculation_layer_counts:
                calculation_layer_counts[calculation_layer] = 0

            calculation_layer_counts[calculation_layer] += 1

        logger.info("\n")

        for layer_type in calculation_layer_counts:

            count = calculation_layer_counts[layer_type]

            logger.info(
                "{} properties were estimated using the {} layer.\n".format(
                    count, layer_type
                )
            )

        logger.info("\n")

        if len(results.exceptions) > 0:

            exceptions = "\n\n".join(str(result) for result in results.exceptions)
            exceptions = exceptions.replace("\\n", "\n")

            # In some cases, an exception will be raised when executing a property but
            # it will not stop the property from being estimated (e.g an error occured
            # while reweighting so a simulation was used to estimate the property
            # instead).
            exceptions_path = os.path.join(
                self.root, self.rundir, "non_fatal_exceptions.txt"
            )

            with open(exceptions_path, "w") as file:
                file.write(exceptions)

            logger.warning(
                "A number of non-fatal exceptions occurred. These were saved to "
                "the %s file." % exceptions_path
            )

        estimated_gradients = {}

        if AGrad is False:
            return results.estimated_properties, estimated_gradients

        jacobian = self._build_pvals_jacobian(mvals)

        # The below snippet will extract any evaluated gradients
        # and map them from gradients with respect to FF parameters,
        # to gradients with respect to FB mathematical parameters.
        for physical_property in results.estimated_properties:

            property_class = physical_property.__class__.__name__

            estimated_gradients[physical_property.id] = np.zeros(
                len(self._gradient_key_mappings)
            )

            for gradient in physical_property.gradients:

                parameter_index = self._gradient_key_mappings[gradient.key]

                gradient_unit = (
                    self._default_units[property_class]
                    / self._parameter_units[gradient.key]
                )

                if isinstance(gradient.value, unit.Quantity):
                    gradient_value = gradient.value.to(gradient_unit).magnitude
                else:
                    gradient_value = gradient.value
                    assert isinstance(gradient_value, float)

                estimated_gradients[physical_property.id][
                    parameter_index
                ] = gradient_value

        for property_id in estimated_gradients:

            pval_gradients = estimated_gradients[property_id]
            mval_gradients = np.matmul(jacobian, pval_gradients)

            estimated_gradients[property_id] = mval_gradients

        return results.estimated_properties, estimated_gradients

    def wq_complete(self):
        """
        Check if all jobs are finished
        This function should have a sleep in it if not finished.

        Returns
        -------
        finished: bool
            True if all jobs are finished, False if not
        """

        estimation_results, _ = self._pending_estimate_request.results()

        return (
            isinstance(estimation_results, EvaluatorException)
            or len(estimation_results.queued_properties) == 0
        )

    def get(self, mvals, AGrad=True, AHess=True):
        """
        Get the objective function value, gradient, and hessian

        Parameters
        ----------
        mvals: np.ndarray
            mvals array containing the math values of the parameters
        AGrad: bool
            Flag for computing gradients of not
        AHess: bool
            Flag for computing hessian or not

        Returns
        -------
        Answer: dict
            Answer = {'X':obj_value, 'G':obj_grad, 'H':obj_hess}
            obj_value: float
            obj_grad: np.ndarray of shape (n_param, )
            obj_hess: np.ndarray of shape (n_param, n_param)

        Notes
        -----
        1. obj_grad is all zero when AGrad == False
        2. obj_hess is all zero when AHess == False or AGrad == False, because the
           hessian estimate depends on gradients
        """

        # Ensure the input flags are actual booleans.
        AGrad = bool(AGrad)
        AHess = bool(AHess)

        # Extract the properties estimated using the unperturbed parameters.
        estimated_data_set, estimated_gradients = self._extract_property_data(
            self._pending_estimate_request, mvals, AGrad
        )

        # compute objective value
        obj_value = 0.0
        obj_grad = np.zeros(self.FF.np)
        obj_hess = np.zeros((self.FF.np, self.FF.np))

        # store details for printing
        self._last_obj_details = {}

        for property_type in self._reference_data_set.property_types:

            self._last_obj_details[property_type] = []

            denominator = (
                self._options.denominators[property_type]
                .to(self._default_units[property_type])
                .magnitude
            )

            weight = self._normalised_weights[property_type]

            for reference_property in self._reference_data_set.properties_by_type(
                property_type
            ):

                reference_value = reference_property.value.to(
                    self._default_units[property_type]
                ).magnitude

                target_property = next(
                    x for x in estimated_data_set if x.id == reference_property.id
                )
                target_value = target_property.value.to(
                    self._default_units[property_type]
                ).magnitude
                target_error = target_property.uncertainty.to(
                    self._default_units[property_type]
                ).magnitude

                diff = target_value - reference_value

                obj_contrib = weight * (diff / denominator) ** 2
                obj_value += obj_contrib

                temperature = reference_property.thermodynamic_state.temperature
                pressure = reference_property.thermodynamic_state.pressure

                self._last_obj_details[property_type].append(
                    (
                        temperature.to(unit.kelvin),
                        pressure.to(unit.atmosphere),
                        target_property.substance.identifier,
                        reference_value,
                        target_value,
                        target_error,
                        diff,
                        weight,
                        denominator,
                        obj_contrib,
                    )
                )

                # compute objective gradient
                if AGrad is True:

                    # get gradients in physical unit
                    grad_array = estimated_gradients[reference_property.id]
                    # compute objective gradient
                    obj_grad += 2.0 * weight * diff * grad_array / denominator ** 2

                    if AHess is True:
                        obj_hess += (
                            2.0
                            * weight
                            * (np.outer(grad_array, grad_array))
                            / denominator ** 2
                        )

        return {"X": obj_value, "G": obj_grad, "H": obj_hess}

    def indicate(self):
        """
        print information into the output file about the last objective function evaluated
        This function should be called after get()
        """
        for property_name, details in self._last_obj_details.items():
            dict_for_print = {
                "  %s %s %s"
                % detail[:3]: "%9.3f %14.3f +- %-7.3f % 7.3f % 9.5f % 9.5f % 9.5f "
                % detail[3:]
                for detail in details
            }
            title = (
                "%s %s\nTemperature  Pressure Substance  Reference  Calculated +- "
                "Stdev     Delta    Weight    Denom     Term  " % (self.name, property_name)
            )
            printcool_dictionary(
                dict_for_print, title=title, bold=True, color=4, keywidth=15
            )
